
import logging
import traceback
import sys
from inspect import Traceback
from logging import LogRecord
from datetime import datetime
from os.path import basename
from typing import List, Optional

import jsonpickle
import json
import os
import threading

# Get log level from env var LOGLEVEL
from rich.console import Console
from rich.logging import RichHandler
from six import raise_from

LOGLEVEL = os.environ.get('LOGLEVEL', 'INFO').upper()

# Global var holding context data
log_context_data = threading.local()

class ThreadingLocalContextFilter(logging.Filter):
    """
    This is a filter which injects contextual information from `threading.local` (log_context_data) into the log.
    """
    def __init__(self, attributes: List[str]):
        super().__init__()
        self.attributes = attributes

    def filter(self, record):
        record.context = ":".join(getattr(log_context_data, a, '-') for a in self.attributes)
        record.file = getattr(log_context_data, "file", None)
        return True

# Wrapper for easy debug function
def wrap(log_f) :
    def f(*args, **kwargs) :
        args = list(args) + list("%s=%s" % (k, str(v)) for k,v in kwargs.items())
        msg = ", ".join(str(arg) for arg in args)
        log_f(msg)
    return f

# Uncaught exception hook
def log_except_hook(*exc_info):
    trace = "".join(traceback.format_exception(*exc_info))
    logger.critical("Unhandled exception. %s", trace)


def obj2json(obj) :
    serialized = jsonpickle.encode(obj)
    return json.dumps(json.loads(serialized), indent=2)

class LogContext(object):
    def __init__(self, network=None, station_id=None, file=None):
        self.context: dict = dict(network=network, station_id=station_id, file=file)

    def __enter__(self):
        for key, val in self.context.items():
            if val :
                setattr(log_context_data, key, val)
        return self

    def __exit__(self, et, ev, tb):
        for key in self.context.keys():
            if self.context[key] and hasattr(log_context_data, key) :
                delattr(log_context_data, key)

        # In case of error, adds context to it
        if ev != None and not isinstance(ev, SystemExit):
            raise_from(Exception(
                "Exception: %s. Context : %s" % (str(ev), str(self.context))), ev)
            return True

class IgnoreAndLogExceptions(object):
    def __init__(self, network=None, station_id=None, file=None):
        pass

    def __enter__(self):
        pass

    def __exit__(self, et, ev, tb):
        if ev is not None:
            if isinstance(ev, KeyboardInterrupt) :
                # Don't capture Ctrl-C
                return False
            else:
                logger.error("Error happened. Ignoring it a continuing. %s", ev)
                logger.exception(ev)
                return True


class RichHandlerContext(RichHandler) :
    """Rich Handler using 'context' atttribute of record as file path """
    def render(
        self,
        *,
        record: LogRecord,
        traceback: Optional[Traceback],
        message_renderable: "ConsoleRenderable",
    ) -> "ConsoleRenderable":
        """Render log for display.

        Args:
            record (LogRecord): logging Record.
            traceback (Optional[Traceback]): Traceback instance or None for no Traceback.
            message_renderable (ConsoleRenderable): Renderable (typically Text) containing log message contents.

        Returns:
            ConsoleRenderable: Renderable to display log.
        """
        level = self.get_level_text(record)
        time_format = None if self.formatter is None else self.formatter.datefmt
        log_time = datetime.fromtimestamp(record.created)

        file=None
        if record.file :
            file = basename(record.file)

        log_renderable = self._log_render(
            self.console,
            [message_renderable] if not traceback else [message_renderable, traceback],
            log_time=log_time,
            time_format=time_format,
            level=level,
            path=file,
            line_no=None,
            link_path=record.pathname if self.enable_link_path else None,
        )
        return log_renderable

# Setup logger
if not sys.stdout.isatty():
    console = Console(
        file=sys.stdout,
        force_terminal=False,
        width=140)
else :
    console = None



rich_handler = RichHandlerContext(omit_repeated_times=False, console=console)
rich_handler.addFilter(ThreadingLocalContextFilter(["network", "station_id"]))
logging.basicConfig(
    level=LOGLEVEL,
    format="{context}\t{message}",
    style="{",
    datefmt="[%X]",
    handlers=[rich_handler])

# Export logger functions
logger = logging.getLogger("rich")
debug = wrap(logger.debug)
info = logger.info
warning = logger.warning
error = logger.error
critical = logger.critical

sys.excepthook = log_except_hook






from collections import defaultdict
from concurrent.futures.thread import ThreadPoolExecutor
from csv import DictReader
from datetime import datetime, timedelta
from typing import Union
from urllib.parse import urlsplit, quote_plus

import numpy as np
from dateutil.relativedelta import relativedelta
from netCDF4 import *
from numpy import timedelta64, datetime64
from numpy.ma import is_masked
from numpy.typing import NDArray
from pandas import DataFrame
import pandas as pd
from pkgutil import get_data
import os
import re
import sys

from six import StringIO

from libinsitu.log import warning

# Name of dimensions and variables
TIME_DIM = 'time'
TIME_VAR = "time"
GLOBAL_VAR = "GHI"
DIFFUSE_VAR = "DHI"
DIRECT_VAR = "BNI"
TEMP_VAR = "T2"
HUMIDITY_VAR = "RH"
PRESSURE_VAR = "P"
WIND_SPEED_VAR = "WS"
WIND_DIRECTION_VAR = "WD"
QC_FLAGS_VAR = "QC"

# Columns for station info, in order of apparition
VALID_COLS = [
    "ID",
    "UID",
    "WMOID",
    "Name",
    "Latitude",
    "Longitude",
    "Elevation",
    "Timezone",
    "StartDate",
    "EndDate",
    "TimeResolution",
    "Address",
    "City",
    "Region",
    "Country",
    "SurfaceType",
    "TopographyType",
    "RuralUrban",
    "Climate",
    "OperationStatus",
 #   "DataBegin",
 #   "DataEnd",
    "ContactName",
    "Institute",
    "Url",
    "CommissionDate",
    "DecommissionDate",
    "DNI_Col",
    "DHI_Col",
    "GHI_Col",
    "QualityStandard",
    "Comment"]

# Variable attributes
VALID_MIN_ATTR = "_valid_min"
VALID_MAX_ATTR = "_valid_max"
FILL_VALUE_ATTR = "_FillValue"

DEFAULT_FILL_VALUE = -999

# Alternate names often found for variables
ALTERNATE_NAMES = {
    DIFFUSE_VAR : ["DIF"],
    DIRECT_VAR : ["DNI"]
}

# Meta data variables
LATITUDE_VAR = "latitude"
LONGITUDE_VAR = "longitude"
ELEVATION_VAR = "elevation"
STATION_NAME_VAR= "station_name"

# Columns no included in Skip_na
COL_NOSKIP=["QC"]

# Global attrs
GLOBAL_TIME_RESOLUTION_ATTR = "time_coverage_resolution"

CLIMATE_ATTRS = [
    "climate",
    "Station_KoeppenGeigerClimate"] # XXX Old convention

STATION_ID_ATTRS = [
    "station_id",
    "Station_Id",  # XXX Old convention
    "id"]

STATION_COUNTRY_ATTRS = [
    "station_country",
    "Station_Country"] # XXX Old conventions

NETWORK_NAME_ATTRS = [
    "Network_Name", # XXX Old convention
    "project"]

# Prefix for global properties
STATION_PREFIX = "Station_"
NETWORK_PREFIX = "Network_"

NA_VALUES=[-999.0, -99.9, -10.0, -9999.0, -99999.0]

CHUNK_SIZE=5000


# List of attributes to search station ID for
STATION_ID_ATTRS = ["Station_ID", "StationInfo_Abbreviation", "station_id", STATION_NAME_VAR]
NETWORK_ID_ATTRS = ["network_id", "Network_ShortName"]

DATA_VARS = [GLOBAL_VAR, DIFFUSE_VAR, DIRECT_VAR, TEMP_VAR, HUMIDITY_VAR, PRESSURE_VAR, WIND_SPEED_VAR, WIND_DIRECTION_VAR]

STATION_INFO_PATTERN = "station-info/%s.csv"
NETWORK_INFO_FILE = "networks.csv"

DATE_FORMAT = '%Y-%m-%d'
TIME_FORMAT_MIN= '%Y-%m-%dT%H:%M'
TIME_FORMAT_SEC= '%Y-%m-%dT%H:%M:%S'

STATION_START_DATA_ATTR = "time_coverage_start"

SECOND = timedelta64(1, 's')
CDL_PATH = "base.cdl"

def parseCSV(res_path, key = "ID") :
    """Generic parser """
    res = dict()
    rows = DictReader(read_res(res_path))
    for row in rows:
        res[row[key]] = {key: parse_value(val) for key, val in row.items()}
    return res

def getStationsInfo(network) :
    """Read station info from CSV"""
    return parseCSV(STATION_INFO_PATTERN % network)

def getNetworksInfo() :
    return parseCSV(NETWORK_INFO_FILE)


def older_than(file1, file2) :
    """Return True if file1 is older than file2"""
    return os.stat(file1).st_mtime < os.stat(file2).st_mtime

def touch(filename):
    """ creates or update the time of a file """
    if os.path.exists(filename):
        os.utime(filename)
    else:
        with open(filename,'a') :
            pass

def getStationInfo(network, station_id) :
    stations = getStationsInfo(network)
    if not station_id in stations :
        raise Exception("Station %s not found in Station Info of %s" % (station_id, network))
    return stations[station_id]

def getNetworkInfo(network) :
    networks = getNetworksInfo()
    if not network in networks :
        raise Exception("Network %s not found in Network infp of %s" % network)
    return networks[network]

def is_uniform(vector) :

    if type(vector) == list :
        vector = np.array(vector)

    if len(vector) <=1 :
        return True
    step = vector[1] - vector[0]
    ref = np.arange(vector[0], vector[-1] + step, step)
    return np.array_equal(ref, vector)

def get_origin_time(ncfile) -> datetime64 :
    timeVar = getTimeVar(ncfile)
    start_time = num2date(0, timeVar.units, timeVar.calendar)
    return np.datetime64(start_time)

def to_int(vals) :
    """Transform single val or array to int """
    if isinstance(vals, np.ndarray):
        return vals.astype(int)
    else:
        return int(vals)

def datetime64_to_sec(ncfile, dates : NDArray[datetime64], unit='s') -> NDArray[int] :
    """Transform datetime64 to number of seconds since origin date """
    origin = get_origin_time(ncfile)
    return to_int((dates - origin) / timedelta64(1, unit))

def str_to_date64(datestr) :
    for format in [TIME_FORMAT_SEC, TIME_FORMAT_MIN, DATE_FORMAT] :
        try :
            return np.datetime64(datetime.strptime(datestr, format))
        except:
            pass

    raise Exception("Unable to parse : " + datestr)


def start_date64(ncfile) :
    if  hasattr(ncfile, STATION_START_DATA_ATTR) :
        return str_to_date64(getattr(ncfile, STATION_START_DATA_ATTR))
    else:
        res = sec_to_datetime64(ncfile, getTimeVar(ncfile)[0])[()]
        warning("No start date set in meta data : taking the first value of ncfile : %s" % res)
        return res

def end_date64(ncfile):
    return sec_to_datetime64(ncfile, getTimeVar(ncfile)[-1])[()]

def seconds_to_idx(ncfile, dates : NDArray[int], ) -> NDArray[int] :
    """Transform seconds since origin to time idx, taking into account resolution and start date"""
    resolution_s = getTimeResolution(ncfile)
    start_sec = datetime64_to_sec(ncfile, start_date64(ncfile))
    return to_int((dates - start_sec) / resolution_s)

def sec_to_datetime64(ncfile, times_s: NDArray[int]) ->  NDArray[datetime64]:
    """Transform number of seconds since start time into datetime64 """
    start_time64 = get_origin_time(ncfile)
    return start_time64 + SECOND * times_s

def read_res(path, encoding="utf8") :
    """Read package resources and returns a fie like object (splitted lines)
    path should be relative to ./res/
    """
    return get_data(__name__, os.path.join("res", path)).decode(encoding).splitlines()

def parse_value(val, split=False) :
    """Parse string value, trying first int, then float. return str value if none are correct"""
    if not isinstance(val, str) :
        return val

    if val is None or val == "":
        return None

    val = val.strip()

    # String
    if val.startswith('"'):
        return val.strip('"')

    # List of things
    if split and "," in val :
        return list(parse_value(item) for item in val.split(","))

    try :
        return int(val)
    except:
        try:
            return float(val)
        except:
            return val


def getTimeResolution(ncfile) :
    """Returns time resolution, in seconds, as saved in meta data"""

    time_var = getTimeVar(ncfile)

    # Formatted as ISO8601 : P10M, P30S, ...
    if hasattr(ncfile, GLOBAL_TIME_RESOLUTION_ATTR) :
        dt = pd.Timedelta(getattr(ncfile, GLOBAL_TIME_RESOLUTION_ATTR))
        return dt.seconds

    if hasattr(time_var, "resolution") :
        # XXX - Support for old versions of NetCDF
        val = time_var.resolution
        val, unit = val.split()
        val = int(val)
        if "min" in unit :
            return val * 60
        elif "sec" in unit:
            return val
        else:
            raise Exception("Unknown unit for time resolution : '%s'" % unit)

    # Guessing from actual first values
    res = int(time_var[1] - time_var[0])
    warning("No resolution set. Guessing :%d seconds" % res)
    return res


def openNetCDF(filename, mode='r', user=None, password=None) :
    """ Open either a filename or OpenDAP URL with user /password"""
    if '://' in filename :
        if user :
            filename = with_auth(filename, user, password)
        filename = "[FillMismatch]" + filename
    return Dataset(filename, mode=mode)

def date_to_timeidx(nc, date) :
    """Transform date to NetCDF index along Time dimension"""
    if isinstance(date, datetime) :
        date = datetime64(date)
    time_sec = datetime64_to_sec(nc, date)
    return seconds_to_idx(nc, time_sec)


def re_pattern(pattern, properties=dict()) :
    """ Transform pattern to regexp matching groups for replacement """
    def subf(match):
        key = match.group(1)
        if key in properties:
            return str(properties[key])
        else:
            if key in ["M", "MM", "YY", "YYYY", "DDD"] :
                pattern = r'\d+' if key == "M" else r'\d' * len(key)
            else:
                pattern = "[A-Z_]+"
            return r'(?P<%s>%s)' % (key, pattern)

    # Transforms pattern to regular expression for matching
    re_pattern = pattern.replace("?", ".").replace("*", ".*")
    return re.sub(r'\{(\w+)\}', subf, re_pattern)

def to_range(year, month=0, day=1) :
    """ Transform a year or month inot a range of datatime """
    yearly=False
    if not month :
        month = 1
        yearly = True

    first = datetime(year, month, day)
    if yearly :
        last = first + relativedelta(years=1)
    else:
        last  = first + relativedelta(months=1)

    return (first, last)


def match_pattern(pattern, value, properties=dict()) :
    """Transforms pattern expression into regexp and extracts the keys """

    reg = re_pattern(pattern, properties)
    match = re.match(reg, value, flags=re.IGNORECASE)

    if not match :
        return False

    res = match.groupdict()

    if "M" in res:
        res["month"] = int(res["M"])
        del res["M"]

    if "MM" in res:
        res["month"]  = int(res["MM"])
        del res["MM"]

    if "DDD" in res:
        res["days"] = int(res["DDD"])
        del res["DDD"]

    if "YYYY" in res:
        res["year"] = int(res["YYYY"])
        del res["YYYY"]

    if "YY" in res:
        year = int(res["YY"])
        res["year"] = year + (2000 if year < 70 else 1900)
        del res["YY"]

    return res

def netcdf_to_dataframe(
        ncfile : Union[Dataset, str],
        start_time: Union[datetime, datetime64]=None,
        end_time:Union[datetime, datetime64]=None,
        rel_start_time=None, # Start time, relative to actual end
        rel_end_time=None, # End time, relative to actual start
        drop_duplicates=True,
        skip_na=False,
        skip_qc=False,
        vars=None,
        user=None,
        password=None,
        chunked=False,
        chunk_size=CHUNK_SIZE,
        steps=1,
        rename_cols=False) :
    """
        Load NETCDF in-situ file (or part of it) into a panda Dataframe, with time as index.


        :param ncfile: NetCDF Dataset or filename, or OpenDAP URL
        :param rename_cols: If True (default) rename solar irradiance columns as per convention (GHI, BNI, DHI)
        :param drop_duplicates: If true (default), duplicate rows are droppped
        :param skip_qc: If True, skip lines with bad QC (at least one failing)
        :param skip_na: If True, drop rows containing only nan values
        :param start_time: Start time (first record by default) : Datetime or datetime64
        :param end_time: End time (last record by default) : Datetile or datetime64
        :param rel_end_time: End time, relative to actual start time : relativedelta
        :param rel_start_time: Start time, relatie to actual end time : relativedelta
        :param vars: List of columns names to convert (all by default)
        :param user: Optional login for OpenDAP URL
        :param password: Optional password OpenDAP URL
        :param chunked: If True, does not load the whole file in memory at once : returns an iterator on Dataframe chunks.
        :param chunk_size: Size of chunks for chunked data
        :param steps: Downsampling (1 by default)
        :return: Pandas Dataframe, or iterator on Dataframes is chunk is activated
        """

    chunks = __nc2df(
        ncfile=ncfile,
        start_time=start_time,
        end_time=end_time,
        rel_start_time=rel_start_time,
        rel_end_time=rel_end_time,
        drop_duplicates=drop_duplicates,
        skip_na=skip_na,
        vars=vars,
        user=user,
        password=password,
        chunked=chunked,
        chunk_size=chunk_size,
        steps=steps,
        rename=rename_cols,
        skip_qc=skip_qc)

    # Handling either single result or chunked generator
    if not chunked :
        for result in chunks:
            return result
    else :
        return chunks


def getTimeVar(nc) :
    for key in nc.variables.keys() :
        if key.lower() == TIME_VAR.lower() :
            return nc.variables[key]
    raise Exception("No time var found")

def __get_attributes(ncfile_or_var) :
    return dict((key, getattr(ncfile_or_var, key)) for key in ncfile_or_var.ncattrs())

def __all_attributes(ncfile) :

    # Global attributes
    attrs = __get_attributes(ncfile)

    # Put single var meta data in global attributes
    attrs[LATITUDE_VAR] = readSingleVar(ncfile, LATITUDE_VAR)
    attrs[LONGITUDE_VAR] = readSingleVar(ncfile, LONGITUDE_VAR)
    attrs[ELEVATION_VAR] = readSingleVar(ncfile, ELEVATION_VAR)
    attrs[STATION_NAME_VAR] = readShortname(ncfile)

    # Add meta data of variables
    attrs["variables"] = dict((varname, __get_attributes(var)) for varname, var in ncfile.variables.items())

    # Put time resolution in seconds in global var
    attrs[GLOBAL_TIME_RESOLUTION_ATTR] = getTimeResolution(ncfile) or 60

    return attrs

def __nc2df(
        ncfile : Union[Dataset, str],
        start_time: Union[datetime, datetime64]=None,
        end_time:Union[datetime, datetime64]=None,
        rel_start_time=None,
        rel_end_time=None,
        drop_duplicates=True,
        skip_na=False,
        skip_qc=False,
        vars=None,
        user=None,
        password=None,
        chunked=False,
        chunk_size=CHUNK_SIZE,
        steps=1, rename=True) :
    """Private generator use by nc2df """

    if isinstance(ncfile, str) :
        ncfile = openNetCDF(ncfile, mode='r', user=user, password=password)

    timeVar = getTimeVar(ncfile)

    size = len(timeVar)

    if rel_start_time is not None :
        start_time = end_date64(ncfile).astype(datetime) + rel_start_time
    if rel_end_time is not None :
        end_time = start_date64(ncfile).astype(datetime) + rel_end_time

    start_idx = max(0, date_to_timeidx(ncfile, start_time)) if start_time else 0
    end_idx = min(date_to_timeidx(ncfile, end_time), size) if end_time else size

    # List of data vars (along time)
    data_vars = []
    for varname, var in ncfile.variables.items() :
        if TIME_DIM in var.dimensions and var != timeVar :
            if vars is None or varname in vars :
                data_vars.append(varname)


    def to_df(start_idx, end_idx) :

        times = sec_to_datetime64(ncfile, timeVar[start_idx:end_idx:steps])

        # Loop on data variables
        data = dict()
        for varname in data_vars :
            var = ncfile.variables[varname]
            floats = var[start_idx:end_idx:steps]

            #if hasattr(var, "least_significant_digit") :
            #    digits = var.least_significant_digit
            #    floats = np.around(floats, decimals=digits)

            data[varname] = floats

        df = DataFrame(data, index=times)

        # Add meta data to attributes of the Dataframe
        df.attrs.update(__all_attributes(ncfile))

        # Drop duplicated : only keep last
        if drop_duplicates :
            df = df[~df.index.duplicated(keep="last")]

        # Drop NA ?
        if skip_na :
            subset = list(col for col in df.columns if col not in COL_NOSKIP)
            df = df.dropna(axis=0, how='all', subset=subset)

        if skip_qc and QC_FLAGS_VAR in df.columns :
            df = df[df[QC_FLAGS_VAR] == 0]

        # Rename variables
        if rename :
            for dest, sources in ALTERNATE_NAMES.items():
                for source in sources :
                    if source in df.columns :
                        warning("Renaming %s -> %s" % (source, dest))
                        df = df.rename(columns={source:dest})


        return df

    if chunked :
        chunk_size = chunk_size * steps
        for idx in range(start_idx, end_idx, chunk_size):
            yield to_df(start_idx=idx, end_idx=min(idx + chunk_size, end_idx))
    else :
        yield to_df(start_idx, end_idx)

def time2str(val, seconds=False) :
    """Format date to the minute """
    if val is None :
        return ""
    if isinstance(val, datetime64) :
        return np.datetime_as_string(val, unit='s' if seconds else 'm')
    elif isinstance(val, datetime) :

        return val.strftime(TIME_FORMAT_SEC if seconds else TIME_FORMAT_MIN)

    raise Exception("Unknown date type : %s" % type(val))

def str2time64(val) :
    if val is None or val == "":
        return None
    return np.datetime64(val)


MIN_STEP=0
MAX_STEP=7200

def get_periods(time_s) :
    """Compute list of periods, by occurrence. Return list of (period, count)"""

    steps = time_s[1:] - time_s[0:len(time_s) - 1]
    unique, counts = np.unique(steps, return_counts=True)

    filtered_unique = unique[(unique > MIN_STEP) & (unique < MAX_STEP)]
    filtered_counts = counts[(unique > MIN_STEP) & (unique < MAX_STEP)]

    indices = np.argsort(-filtered_counts)[:3]
    periods_dic =  dict((filtered_unique[idx], filtered_counts[idx]) for idx in indices)
    return list((int(period), count) for period, count in periods_dic.items())


def parseTimezone(val) :
    """Parse timezone UTC+HH:MM to timedelta"""
    val = val.strip("UTC")

    if ":" in val :
        hh, mm = val.split(":")
        hh= int(hh)
        mm = int(mm)
    else :
        hh = int(val)
        mm = 0

    if hh < 0 :
        mm = -mm

    return pd.to_timedelta(60*hh+mm, "min")

def with_auth(url, user, password) :
    parts = urlsplit(url)
    return "%s://%s:%s@%s/%s" % (parts.scheme, quote_plus(user), quote_plus(password), parts.netloc, parts.path)


def fill_str(nc, varname, shortname) :

    var = nc.variables[varname]

    if var.dtype == str :

        # Variable length string
        var[0] = shortname
    else:

        dim = var.dimensions[0]
        size = nc.dimensions[dim].size

        # Transform to null terminated fixed length array of chars
        shortname_ = stringtochar(np.array(shortname, 'S%d' % size))
        var[:] = shortname_

def read_str(var) :

    if var.dtype == str :
        # Var length string
        return var[0]

    # Fixed char array
    char_array = var[:]

    # This is a 0D (no dimension) array !
    string_array = chartostring(char_array)

    return string_array[()]

def readSingleVar(nc, var) :
    """Read a meta variable encoded as a single value variable """
    arr = nc.variables[var][:].flatten()
    if len(arr) == 0 :
        return None
    else:
        return arr[0]

def readShortname(ncfile) :
    # Then read the content of the dedicated var
    return read_str(ncfile.variables[STATION_NAME_VAR])

def getMult(attrs, keys):
    """Get value of dict or attribute in NCfile, looking several possbilities"""

    if isinstance(attrs, Dataset) :
        attrs  = dict((attr, getattr(attrs, attr)) for attr in attrs.ncattrs())

    for name in keys:
        if name in attrs:
            return attrs[name]
    return None

def getNetworkId(attributes_or_ncfile) :
    return getMult(attributes_or_ncfile, NETWORK_ID_ATTRS)

def getStationId(attributes_or_ncfile) :
    return getMult(attributes_or_ncfile, STATION_ID_ATTRS)

def getProperties(network_id, station_id) :
    """Gather Network_ and Station_ properties """

    # Get properties for this station
    properties = {STATION_PREFIX + k : v for k, v in getStationInfo(network_id, station_id).items()}

    # Add properties of this network
    for key, val in getNetworkInfo(network_id).items():
        properties[NETWORK_PREFIX + key] = val

    return properties


def qc_masks(df) :
    """Parse metadata of a QC bitmap and returns dict of meaning => mask"""
    attrs = df.attrs["variables"][QC_FLAGS_VAR]
    return {meaning: mask for meaning, mask in zip(
        attrs["flag_meanings"].split(),
        attrs["flag_masks"]
    )}

def parse_bool(value) :
    return value in ["true", "True", "1", "yes", "Yes"]

def parallel_map(fn, iterable, parallel, max_workers=None) :
    """Helper util to map either seuquentially or in parallel """
    if parallel :
        exec = ThreadPoolExecutor(max_workers=max_workers)
        return exec.map(fn, iterable)
    else:
        return map(fn, iterable)

def df_to_csv(df, out=sys.stdout, **args) :
    """Output CSV to stdout"""
    output = StringIO()
    df.to_csv(output, **args)
    output.seek(0)
    out.write(output.read())

def df_to_json(df, out=sys.stdout, **args) :
    """Output CSV to stdout"""
    output = StringIO()
    df.to_json(output, **args)
    output.seek(0)
    out.write(output.read())

class DefaultDict(defaultdict) :
    """Default awnsering 'True' to X in dict """

    def __contains__(self, item):
        return True



import numpy as np

class BaseGraphs :
    """ Base class holding source data to compute graph. Implementations inherit from it """

    def __init__(self,
            meas_df,
            sp_df,
            flag_df,
            cams_df,
            stat_test,
            horizons,
            latitude,
            longitude,
            elevation,
            station_id="-",
            station_name="-",
            show_flag=-1):

        """
         ShowFlag=-1     : only show non-flagged data
         ShowFlag=0      : show all data without filtering nor tagging flagged data
         ShowFlag=1      : show all data and highlight flagged data in red
        """

        self.meas_df = meas_df
        self.sp_df = sp_df
        self.show_flag=show_flag
        self.cams_df = cams_df
        self.stat_test = stat_test
        self.horizons = horizons
        self.latitude = latitude
        self.longitude = longitude
        self.elevation = elevation
        self.station_id = station_id
        self.show_flag = show_flag
        self.station_name = station_name

        if show_flag == -1:
            # Hide all data with having at least one QC error
            meas_df.loc[
                flag_df.QCfinal != 0,
                ["GHI", "DHI", "BNI"]] = np.nan

        # Aliases
        self.GHI = meas_df.GHI
        self.DIF = meas_df.DHI
        self.DNI = meas_df.BNI

        self.TOA = sp_df.TOA
        self.TOANI = sp_df.TOANI
        self.GAMMA_S0 = sp_df.GAMMA_S0
        self.THETA_Z = sp_df.THETA_Z
        self.ALPHA_S = sp_df.ALPHA_S
        self.SZA = sp_df.SZA
        self.flag_df = flag_df
        self.QCfinal = flag_df.QCfinal

        self.GHI_est = self.DIF + self.DNI * np.cos(self.THETA_Z)


from pvlib.iotools.midc import MIDC_VARIABLE_MAP, TZ_MAP
from libinsitu.common import GLOBAL_VAR, DIRECT_VAR, DIFFUSE_VAR, TEMP_VAR, HUMIDITY_VAR, PRESSURE_VAR, WIND_SPEED_VAR, \
    WIND_DIRECTION_VAR, NA_VALUES
from libinsitu.handlers.base_handler import InSituHandler, ZERO_DEG_K
import pandas as pd
from datetime import datetime, timedelta

from libinsitu.log import warning, info

VARIABLE_MAP = {
    'ghi' : GLOBAL_VAR,
    'dni' : DIRECT_VAR,
    'dhi' : DIFFUSE_VAR,
    'wind_speed' : WIND_SPEED_VAR,
    'temp_air' : TEMP_VAR,
    'relative_humidity': HUMIDITY_VAR}

TIME_FORMAT="%Y%j%H%M" # YYYYJJJHHMM
ONE_DAY = timedelta(days=1)

class NRELHandler(InSituHandler) :

    def _read_chunk(self, stream) :

        # Check file is not Error from REST API
        first_line = stream.readline()
        if first_line.startswith("Error") :
            return None

        stream.seek(0)

        # Count expected cols
        first_row = pd.read_csv(stream, nrows=0)
        col_length = len(first_row.columns)
        stream.seek(0)

        # CSV to pandas
        data = pd.read_csv(stream, usecols=range(col_length), na_values=NA_VALUES)
        data = format_index_raw(data).tz_convert("UTC")

        station_id = self.properties["Station_ID"]

        if not station_id in MIDC_VARIABLE_MAP :
            raise Exception("Station %s not defined in variable maps (%s)" % (station_id, list(MIDC_VARIABLE_MAP.keys())))

        # Transform variable map to our variable names
        mapping = MIDC_VARIABLE_MAP[station_id]
        mapping = {key: VARIABLE_MAP[val] for key, val in mapping.items()}

        # Filter only the columns that we can find
        columns = list(data.columns)
        for key in list(mapping.keys()) :
            if not key in columns :
                warning("Column %s was not found in dataset", key)
                del mapping[key]

        #info("Columns : %s", columns)

        # Filter and rename columns
        data = data[list(mapping.keys())]
        data = data.rename(columns=mapping)

        columns = list(data.columns)

        if TEMP_VAR in data.columns :
            data[TEMP_VAR] = data[TEMP_VAR] + ZERO_DEG_K # T2: °C -> K

        if HUMIDITY_VAR in columns :
            data[HUMIDITY_VAR] = data[HUMIDITY_VAR] / 100  # percent -> 1

        return data


    def data_vars(self):
        """ @override """
        return [GLOBAL_VAR, DIFFUSE_VAR, DIRECT_VAR, TEMP_VAR, HUMIDITY_VAR, PRESSURE_VAR, WIND_SPEED_VAR, WIND_DIRECTION_VAR]


def parseTime(yyyyjjjhhmm) :
    """ Parse time, handling corner case of H:24 """
    hh = yyyyjjjhhmm[7:9]
    plusDay = 0
    if hh == "24" :
        yyyyjjjhhmm = yyyyjjjhhmm[:7] + "00" + yyyyjjjhhmm[9:]
        plusDay = 1
    res = datetime.strptime(yyyyjjjhhmm, TIME_FORMAT)
    if plusDay > 0 :
        res += ONE_DAY
    return res


def tzCol(data) :
    tz_columns = list(str(col) for col in data.columns if len(str(col)) == 3 and str(col)[1:] == "ST")
    if len(tz_columns) > 1:
        raise Exception("Found more than 1 potential Timezone columns : %s" % str(tz_columns))
    if len(tz_columns) == 0:
        raise Exception("Found no  Timezone column")
    return tz_columns[0]

# Adapation of function from pvlib / midc to handle hour=24
def format_index_raw(data):
    """Create DatetimeIndex for the Dataframe localized to the timezone provided
    as the label of the third column.

    Parameters
    ----------
    data: Dataframe
        Must contain columns 'Year' and 'DOY'. Timezone must be found as the
        label of the third (time) column.

    Returns
    -------
    data: Dataframe
        The data with a Datetime index localized to the provided timezone.
    """


    # There shoulf be a single TZ column
    tz_raw = tzCol(data)

    timezone = TZ_MAP.get(tz_raw, tz_raw)
    year = data.Year.apply(str)
    jday = data.DOY.apply(lambda x: '{:03d}'.format(x))
    time = data[tz_raw].apply(lambda x: '{:04d}'.format(x))
    time_str = year + jday + time
    index = time_str.apply(parseTime)
    data = data.set_index(index)
    data = data.tz_localize(timezone)
    return data


from libinsitu.handlers import SAURAN, SKYNET
from libinsitu.handlers.ABOM import ABOMHandler
from libinsitu.handlers.BSRN import BSRNHandler
from libinsitu.handlers.ESMAP import ESMAPHandler
from libinsitu.handlers.IEA_PVPS import IEA_PVPSHandler
from libinsitu.handlers.ISE_PVLive import ISEPVLive
from libinsitu.handlers.NREL_MIDC import NRELHandler
from libinsitu.handlers.RAD import RADHandler
from libinsitu.handlers.SAURAN import SAURANHandler
from libinsitu.handlers.SKYNET import SkyNetHandler
from libinsitu.handlers.enerMENA import EnerMENAHandler
from libinsitu.handlers.base_handler import InSituHandler

# Static map of handlers
HANDLERS = {
    "BSRN" : BSRNHandler,
    "enerMENA" : EnerMENAHandler,
    "ABOM" : ABOMHandler,
    "ESMAP": ESMAPHandler,
    "ISE_PVLive" : ISEPVLive,
    "SAURAN" : SAURANHandler,
    "NREL_MIDC" : NRELHandler,
    "SURFRAD" : RADHandler,
    "SOLRAD" : RADHandler,
    "SKYNET" : SkyNetHandler,
    "IEA_PVPS" : IEA_PVPSHandler
}

def listNetworks() :
    return sorted(list(HANDLERS.keys()))




import copy

import matplotlib as mpl
from matplotlib import dates as mdates, pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.pyplot import gca
from pandas import DataFrame
from pvlib.clearsky import detect_clearsky

from libinsitu import info, CLIMATE_ATTRS, STATION_COUNTRY_ATTRS, NETWORK_NAME_ATTRS
from matplotlib import cm
import numpy as np
from matplotlib.colors import ListedColormap
from libinsitu.qc.base_graphs import BaseGraphs


NB_MIN_IN_DAY = 24 * 60
FONT_SIZE = 8
MC_CLEAR_COLOR = 'mediumseagreen'


def makeCustomColormap(NWhite=2, ColorGrey=0.8, NGrey=25, cmColor='viridis', NColor=100):

    cmpGrey = ColorGrey * np.ones((1, 3))
    cmpColor = cm.get_cmap(cmColor, 256)(np.linspace(0, 1, NColor + 10))[10:, :]
    M0 = np.hstack((np.ones((NGrey, 1)) @ cmpGrey + (np.expand_dims((np.linspace(0, 1, NGrey)), axis=0).T) @ (
                cmpColor[0, 0:3] - cmpGrey), np.ones((NGrey, 1))))
    cmWGC = ListedColormap(np.vstack((np.ones((NWhite, 4)), M0, cmpColor)), name='jet_ymsd')
    return cmWGC

COLORMAP_PLOTS = makeCustomColormap(NWhite=2, ColorGrey=0.8, NGrey=25, cmColor='cividis', NColor=100)
COLORMAP_DENSITY = makeCustomColormap(NWhite=1, ColorGrey=0.8, NGrey=50, cmColor='viridis', NColor=200)
COLORMAP_SHADING = makeCustomColormap(NWhite=2, ColorGrey=0.8, NGrey=25, cmColor='cividis', NColor=100)

def conv2(v1, v2, m, mode='same'):
    tmp = np.apply_along_axis(np.convolve, 0, m, v1, mode)
    return np.apply_along_axis(np.convolve, 1, tmp, v2, mode)


class BaseMatplotlibGraphs(BaseGraphs):

    def plot_timeseries(self, label, data, ymax) :

        info("plotting timeseries for %s" % label)

        idxPlot = (self.TOA > 0) & (data.values > -50)

        index = data.index

        axe = gca()
        axe.plot(
            data.index[idxPlot],
            data.values[idxPlot],
            color='b',
            alpha=0.8,
            label='meas.',
            lw=0.2)

        plt.ylim((0, ymax))
        plt.xlim((
            index.values[0],
            index.values[-1]))

        axe.set_ylabel(label + " (W/m2)", size=8)
        plt.setp(axe.get_xticklabels(), visible=False)

    def plot_heatmap_timeseries(self, label, data, cmax):

        info("plotting heatmap timeseries for %s " % label)

        index = data.index
        axe = gca()

        nb_days = np.int64(len(index) / NB_MIN_IN_DAY)
        values = copy.deepcopy(data.values)
        M2D = np.reshape(values, (nb_days, NB_MIN_IN_DAY)).T
        deltaT = int(np.round(self.longitude / 360 * 24 * 60))
        M2D2 = np.roll(M2D, deltaT, axis=0)
        M2D2[M2D2 < -100] = np.nan

        x_min, x_max = mdates.date2num([index[0].date(), index[-1].date()])

        im00 = axe.imshow(
            M2D2,
            extent=[x_min, x_max, 24, 0],
            aspect='auto', cmap=COLORMAP_PLOTS, alpha=1)

        axe.xaxis_date()

        plt.setp(axe.get_xticklabels(), visible=False)
        axe.set_yticks(np.arange(0, 23, 6))
        axe.set_ylabel('Time of the day', fontsize=FONT_SIZE)

        # Plot sunrise and sunset
        def plot_limit(limit) :
            h_lt = limit + float(deltaT) / 60
            h_lt[h_lt > 24] = h_lt[h_lt > 24] - 24
            h_lt[h_lt < 0] = h_lt[h_lt < 0] + 24
            axe.plot(mdates.date2num(index), h_lt, 'k--', linewidth=0.75, alpha=0.8)

        plot_limit(self.sp_df.SR_h) # Sunrise
        plot_limit(self.sp_df.SS_h) # Sunset

        im00.set_clim(0, cmax)
        axe.text(mdates.date2num(index)[0] + 5, 21, label, size=10)

        mpl.rcParams['ytick.labelsize'] = FONT_SIZE
        plt.xlim((index.values[0], index.values[-1]))
        plt.ylim((0, 24))

        if self.show_flag == 1:

            timeLMT = index.values + np.timedelta64(int(self.longitude / 360 * 24 * 60 * 60), 's')
            day = timeLMT.astype('datetime64[D]').astype(index.values.dtype)
            TOD = 1 + (timeLMT - day).astype('timedelta64[s]').astype('double') / 60 / 60

            plt.plot(day[self.QCfinal], TOD[self.QCfinal], 'rs', markersize=1, alpha=0.8, label='flag')

            axe.legend(loc='lower right')


    def plot_ratio_heatmap(self, ratios, filter, h1, h2, ylimit, title, y_label, Ratio4C=0.5, bg_color=None, hlines=[]) :

        axe = gca()
        axe.set_yticks(np.arange(0.2, 2, 0.1))

        index = ratios.index

        Filteridx = filter & (ratios.values > 0) & (ratios.values < 100) & (self.TOA > 0)

        x = mdates.date2num(ratios.index[Filteridx])
        y = ratios.values[Filteridx]

        if len(x) == 0 :
            return

        hist, xedges, yedges = np.histogram2d(x, y, bins=[int((x[-1] - x[0])), 400],
                                              range=[[x[0], x[-1]], [0.25, 1.75]])
        if int((x[-1] - x[0])) > 10 * len(h2):
            hist = conv2(h1, h2, hist)
        yedges, xedges = np.meshgrid(0.5 * (yedges[:-1] + yedges[1:]), 0.5 * (xedges[:-1] + xedges[1:]))
        im00 = plt.scatter(xedges.flatten(), yedges.flatten(), s=3, c=hist.flatten(), cmap=COLORMAP_DENSITY)
        im00.set_clim(0, Ratio4C * max(hist.flatten()))


        plt.plot(ratios.index, np.ones(len(ratios)), 'r--', alpha=0.5)

        plt.ylim((1 - ylimit, 1 + ylimit))

        axe.set_ylabel(y_label, fontsize=FONT_SIZE)
        plt.xlim((index.values[0], index.values[-1]))
        mpl.rcParams['xtick.labelsize'] = FONT_SIZE
        mpl.rcParams['ytick.labelsize'] = FONT_SIZE
        axe.xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m'))
        axe.xaxis_date()

        for (delta_y, width) in hlines :
            plt.plot([ratios.index[0], ratios.index[-1]], [1-delta_y, 1-delta_y], 'k-.', alpha=0.4, linewidth=width)
            plt.plot([ratios.index[0], ratios.index[-1]], [1+delta_y, 1+delta_y], 'k-.', alpha=0.4, linewidth=width)

        if self.show_flag == 1:
            plt.plot(mdates.date2num(ratios.index[self.QCfinal]),
                     ratios.values[self.QCfinal], 'rs', markersize=1, alpha=0.8, label='flag')
            axe.legend(loc='lower right')

        if bg_color:

            axe.xaxis.label.set_color(bg_color)  # setting up X-axis label color to yellow
            axe.yaxis.label.set_color(bg_color)  # setting up Y-axis label color to blue

            axe.tick_params(axis='y', colors=bg_color)  # setting up Y-axis tick color to black

            axe.spines['left'].set_color(bg_color)  # setting up Y-axis tick color to red
            axe.spines['right'].set_color(bg_color)
            axe.spines['top'].set_color(bg_color)  # setting up above X-axis tick color to red
            axe.spines['bottom'].set_color(bg_color)
            axe.text(mdates.date2num(index.values[0]) + 10, 1 + ylimit * 0.75, title, color=bg_color)

        else:
            axe.text(mdates.date2num(index.values[0]) + 10, 1 + ylimit * 0.75, title)

        return axe


    def generic_qc_graph(self, x, y, xlabel, ylabel, xrange, yrange, legend, lines, clim_ratio=0.25) :
        """ Generic function to display QC heat map with limits """

        info("Plotting QC test: %s" % legend)

        ax = plt.gca()
        plt.text(0.01, 0.9, legend, transform=ax.transAxes)

        hist, xedges, yedges = np.histogram2d(
            x=x,
            y=y,
            bins=[200, 200],
            range=[xrange, yrange])

        yedges, xedges = np.meshgrid(
            0.5 * (yedges[:-1] + yedges[1:]),
            0.5 * (xedges[:-1] + xedges[1:]))

        im = plt.scatter(
            xedges.flatten(),
            yedges.flatten(), s=3,
            c=hist.flatten(),
            cmap=COLORMAP_DENSITY)

        im.set_clim(0, clim_ratio * max(hist.flatten()))

        for x, y in lines :
            plt.plot(x, y, 'k--', alpha=0.4, linewidth=0.8)

        plt.xlabel(xlabel, fontsize=FONT_SIZE)
        plt.ylabel(ylabel, fontsize=FONT_SIZE)

        plt.xlim(xrange)
        plt.ylim([
            yrange[0],
            yrange[1] * 1.2])

        return im

    def plot_bsrn_1c(self, component, component_name, limits) :

        legend= 'BSRN 1C ' + component_name + ": {:.2f}% / {:.2f}%".format(
            self.stat_test['T1C_ppl_' + component_name],
            self.stat_test['T1C_erl_' + component_name])

        # XXX should be done beforehand
        filter = (self.TOA > 0) & (component > 0) & (component < 2000) & (self.TOA > 0) & (self.TOA < 2000)

        x = self.TOA[filter]
        y = component[filter]

        TOANI = self.TOANI[filter]
        GAMMA_S0 = self.GAMMA_S0[filter]

        # Draw limits
        limits_xy = []
        for a, b, c in limits:
            yy = a * TOANI * np.sin(GAMMA_S0) ** b + c

            # Poly appromimation
            tx = np.arange(min(x), max(x), 100)
            fpoly = np.poly1d(np.polyfit(x, yy, 5))

            limits_xy.append([tx, fpoly(tx)])

        self.generic_qc_graph(
            x=x, y=y,
            xlabel='Top of atmosphere (TOA) (W/m2)',
            ylabel=component_name + "W/m2",
            xrange=[1, 1300],
            yrange=[0, 1400],
            legend=legend,
            lines=limits_xy,
            clim_ratio=0.25)

        #if ShowFlag == 1:
        #    plt.plot(x[flag_df['T1C_erl_' + PrmYi[jj]]], y[flag_df['T1C_erl_' + PrmYi[jj]]], 'rs',
        #             markersize=1, alpha=0.5)
        #    plt.plot(x[flag_df['T1C_ppl_' + PrmYi[jj]]], y[flag_df['T1C_ppl_' + PrmYi[jj]]], 'rs',
        #             markersize=1, alpha=0.5, label='erl')
        #    plt.legend(loc='lower right')

    def bsrn_2c(self):

        filter = (self.GHI > 50) & (self.SZA < 90)

        line = [
            [0, 75, 75, 100],
            [1.05, 1.05, 1.1, 1.1]]

        return self.generic_qc_graph(
            legend="BSRN-2C : {:.2f}% ".format(self.stat_test['T2C_bsrn_kt']),
            x=self.SZA[filter], xlabel='Solar zenith angle (°)', xrange=[10, 95],
            y=self.flag_df.K[filter], ylabel='DIF/GHI (-)', yrange = [0, 1.25],
            lines=[line],
            clim_ratio=0.8)

    def seri_kn(self) :

        filter = (self.DNI > 0) & (self.GHI > 0) & (self.SZA < 90)

        line = [
            [0, 0.8, 1.35, 1.35],
            [0, 0.8, 0.8, 0]]

        return self.generic_qc_graph(
            legend = "SERI-kn : {:.2f}% ".format(self.stat_test['T2C_seri_knkt']),
            x=self.flag_df.KT[filter], xlabel='GHI/TOA (-)', xrange=(0, 1.5),
            y=self.flag_df.Kn[filter], ylabel='DNI/TOANI (-)', yrange=(0, 0.8),
            lines=[line], clim_ratio=0.1)

    def seri_k(self) :

        filter = (self.DIF > 0) & (self.GHI > 0) & (self.SZA < 90)

        line = (
            [0, 0.6, 0.6, 1.35, 1.35],
            [1.1, 1.1, 0.95, 0.95, 0])

        return self.generic_qc_graph(
            legend="SERI-K : {:.2f}% ".format(self.stat_test['T2C_seri_kkt']),
            x=self.flag_df.KT[filter], xlabel='GHI/TOA (-)', xrange=(0, 1.5),
            y=self.flag_df.K[filter], ylabel='DIF/GHI (-)', yrange=(0, 1.4),
            lines=[line],
            clim_ratio=0.1)

    def bsrn_closure(self) :

        filter = (self.DIF > 0) & (self.GHI > 50) & (self.SZA < 90)

        # 4 diagonal lines
        lines = []
        for r in [0.85, 0.92, 1.08, 1.15] :
            lines.append([
                [0.0, 1400.0],
                [0.0, 1400.0 *r]
            ])

        return self.generic_qc_graph(
            legend="BSRN closure : {:.2f}% ".format(self.stat_test['T3C_bsrn']),
            x=self.GHI[filter], xlabel='GHI (W/m2)', xrange=(0, 1400),
            y=self.GHI_est[filter], ylabel='DIF+DNI*CSZA (W/m2)', yrange=(0, 1300),
            lines=lines,
            clim_ratio=0.1)

    def bsrn_closure_ratio(self) :

        filter = (self.DIF > 0) & (self.GHI > 50) & (self.SZA < 90)

        line = [
            [10, 75, 75, 90, 90, 75, 75, 10],
            [1.08, 1.08, 1.15, 1.15, 0.85, 0.85, 0.92, 0.92]]

        return self.generic_qc_graph(
            legend="BSRN closure: {:.2f}% ".format(self.stat_test['T3C_bsrn']),
            x=self.SZA[filter], xlabel='Solar zenith angle (°)', xrange=(0, 100),
            y=self.GHI[filter] / self.GHI_est[filter], ylabel='GHI/(DIF+DNI*CSZA) (-)', yrange=(0.6, 1.2),
            lines=[line],
            clim_ratio=0.5)


    def print_info(self) :

        Y0 = 0.90
        dY = 0.15
        gs0 = GridSpec(9, 12)
        gs0.update(left=0.015, right=0.99, bottom=0.05, top=0.99, hspace=0.01, wspace=0.05)

        climate = _get_meta(self.meas_df, CLIMATE_ATTRS)
        country = _get_meta(self.meas_df, STATION_COUNTRY_ATTRS)
        source = _get_meta(self.meas_df, NETWORK_NAME_ATTRS)

        CodeInfo = {
            "project": "CAMS2-73",
            "author": 'ARMINES, DLR',
            "name": 'libinsitu - Visual plausibility control',
            "vers": get_version()}

        NbDays = len(self.meas_df.index[self.GHI > 0].normalize().unique())
        AvgGHI = sum(self.GHI[self.GHI > 0]) * 1 / 60 / NbDays * 365 / 1000
        AvgDHI = sum(self.DIF[self.DIF > 0]) * 1 / 60 / NbDays * 365 / 1000
        AvgDNI = sum(self.DNI[self.DNI > 0]) * 1 / 60 / NbDays * 365 / 1000
        AvailGHI = sum((self.GHI > -2) & (self.TOA > 0)) / sum((self.TOA > 0)) * 100
        AvailDHI = sum((self.DIF > -2) & (self.TOA > 0)) / sum(self.TOA > 0) * 100
        AvailDNI = sum((self.DNI > -2) & (self.TOA > 0)) / sum(self.TOA > 0) * 100

        DateStrStart = self.meas_df.index[self.GHI > 0][0].strftime("%Y-%m-%d")
        DateStrEnd = self.meas_df.index[self.GHI > 0][-1].strftime("%Y-%m-%d")

        ax01 = plt.subplot(gs0[0, 8])
        ax01.text(0.01, Y0 - 0 * dY, 'Source: ' + source, size=FONT_SIZE)
        ax01.text(0.01, Y0 - 1 * dY, self.station_id + ': ' + self.station_name, size=FONT_SIZE)  # 'ID/ Station'
        ax01.text(0.01, Y0 - 2 * dY, "latitude: {:.2f}°".format(self.latitude), size=FONT_SIZE)
        ax01.text(0.01, Y0 - 3 * dY, "longitude: {:.2f}°".format(self.longitude), size=FONT_SIZE)
        ax01.text(0.01, Y0 - 4 * dY, "altitude: {:.0f}m".format(self.elevation), size=FONT_SIZE)
        ax01.text(0.01, Y0 - 5 * dY, "country: {} ".format(country), size=FONT_SIZE)
        ax01.text(0.01, Y0 - 6 * dY, "Köppen-Geiger climate: {}".format(climate), size=FONT_SIZE)
        ax01.axis('off')

        ax02 = plt.subplot(gs0[0, 10])
        ax02.text(0.01, Y0 - 1 * dY, '.         Period:  {} - {}'.format(DateStrStart, DateStrEnd), size=FONT_SIZE)
        ax02.text(0.01, Y0 - 2 * dY, '  Annual sums:', size=FONT_SIZE)
        ax02.text(0.01, Y0 - 3 * dY, 'GHI:   {0:.0f} kWh/m2'.format(AvgGHI), size=FONT_SIZE)
        ax02.text(0.01, Y0 - 4 * dY, 'DIF:   {0:.0f} kWh/m2'.format(AvgDHI), size=FONT_SIZE)
        ax02.text(0.01, Y0 - 5 * dY, 'DNI:   {0:.0f} kWh/m2'.format(AvgDNI), size=FONT_SIZE)
        ax02.text(0.01, Y0 - 6 * dY, CodeInfo["name"] + ' ' + CodeInfo["vers"] + '', size=FONT_SIZE)
        ax02.axis('off')

        ax03 = plt.subplot(gs0[0, 11])
        ax03.text(0.01, Y0 - 2 * dY, '        Days of data: {}'.format(NbDays), size=FONT_SIZE)
        ax03.text(0.01, Y0 - 3 * dY, '      ({0:.1f}% availability)'.format(AvailGHI), size=FONT_SIZE)
        ax03.text(0.01, Y0 - 4 * dY, '      ({0:.1f}% availability)'.format(AvailDHI), size=FONT_SIZE)
        ax03.text(0.01, Y0 - 5 * dY, '      ({0:.1f}% availability)'.format(AvailDNI), size=FONT_SIZE)
        ax03.axis('off')

    def histo_qc(self, comp, x, x_label, legend_pos=None, y_label=False) :

        axe = gca()

        idxPlot_qc = (comp > 5) & (self.SZA < 90) & (self.QCfinal == 0)
        idxPlot_all = (comp > 5) & (self.SZA < 90)
        hist_qc, xedges = np.histogram(x[idxPlot_qc], bins=500, range=[0, 1.2])
        hist_all, xedges = np.histogram(x[idxPlot_all], bins=500, range=[0, 1.2])
        xval = (xedges[1:] + xedges[:-1]) / 2
        axe.fill_between(xval, 0 * xval, hist_all, color=[204 / 255, 0 / 255, 0 / 255], label='flagged data')
        axe.fill_between(xval, 0 * xval, hist_qc, color=[102 / 255, 178 / 255, 255 / 255], label='valid data')

        if legend_pos :
            axe.legend(loc=legend_pos)

        if y_label :
            axe.set_ylabel('count')

        axe.set_xlim([0, 1.1])
        axe.set_xticks([0, 0.2, 0.4, 0.6, 0.8, 1.0])
        axe.set_xticklabels([0, 0.2, 0.4, 0.6, 0.8, 1])
        axe.set_xlabel(x_label)
        axe.set_yticks([])

    def horizontality_graph(self):

        axe = gca()

        # Aliases
        CLEAR_SKY_GHI = self.cams_df.CLEAR_SKY_GHI
        CLEAR_SKY_DNI = self.cams_df.CLEAR_SKY_DNI

        isClearSky = detect_clearsky(self.GHI, CLEAR_SKY_GHI)

        YYL = [0.8, 1.2]

        idxPlot = isClearSky & (self.DIF > 0) & (self.GHI > 100) & (self.SZA < 90) & \
                      (CLEAR_SKY_GHI.values > 50) & (self.GHI.values > 50) & (CLEAR_SKY_DNI.values > 0)

        vSAA = self.ALPHA_S.values
        if self.latitude < 0:
            vSAA[vSAA * 180 / np.pi > 180] = vSAA[vSAA * 180 / np.pi > 180] - 2 * np.pi

        if (self.latitude < 0):
            angle_filter = np.abs(vSAA * 180 / np.pi) < 60
        else:
            angle_filter = np.abs(vSAA * 180 / np.pi - 180) < 60

        kc = self.GHI[idxPlot & angle_filter & (CLEAR_SKY_DNI.values > 0)] / \
             CLEAR_SKY_GHI[idxPlot & angle_filter & (CLEAR_SKY_DNI.values > 0)]

        S = kc.resample('1D', label='left').sum()
        C = kc.resample('1D', label='left').count()

        Dailydata = DataFrame({'Avgkc': S[C > 30] / C[C > 30]}, index=C.index)

        data4plot = DataFrame(
            {'kc': self.GHI[idxPlot] / CLEAR_SKY_GHI[idxPlot], \
             'SAA': vSAA[idxPlot] * 180 / np.pi, \
             'day': self.meas_df[idxPlot].index.floor(freq='D')}, \
            index=self.meas_df[idxPlot].index)
        data4plot = data4plot.join(Dailydata, on='day', how='left')
        ix = data4plot.Avgkc > 0
        xPlot = data4plot.SAA[ix].values
        yPlot = data4plot.kc[ix].values / data4plot.Avgkc[ix].values

        dxx = 0 if self.latitude >= 0 else 180

        hist, xedges, yedges = np.histogram2d(x=xPlot, y=yPlot, bins=[180, 100], range=[[0 - dxx, 360 - dxx], YYL])
        plt.plot([0 - dxx, 360 - dxx], [1, 1], 'r--', alpha=0.4, linewidth=0.8)
        plt.xlim((0 - dxx, 360 - dxx))
        axe.text(5 - dxx, 0.97 * YYL[1], 'Test of the horizontality of the GHI sensor')

        yedges, xedges = np.meshgrid(0.5 * (yedges[:-1] + yedges[1:]), 0.5 * (xedges[:-1] + xedges[1:]))

        im00 = plt.scatter(xedges[hist > 0], yedges[hist > 0], s=3, c=hist[hist > 0], cmap=COLORMAP_DENSITY)
        im00.set_clim(0, 0.7 * max(hist.flatten()))

        plt.ylabel('kc/kc_daily (-)', fontsize=FONT_SIZE)
        plt.xlabel('Solar azimuth angle (°)', fontsize=FONT_SIZE)

        axe.set_ylim(YYL)

        plt.colorbar(im00, label='point density (-)')

    def shadow_analysis(self, label, comp, ref, cmax) :

        axe = gca()

        idxSC = (self.GAMMA_S0 > 1 / 50)
        vSEA = self.GAMMA_S0[idxSC]
        vSAA = self.ALPHA_S[idxSC]
        if self.latitude < 0:
            vSAA[vSAA * 180 / np.pi > 180] = vSAA[vSAA * 180 / np.pi > 180] - 2 * np.pi

        SELMax = 45

        vK = comp[idxSC] / ref[idxSC]
        idx_sort = np.argsort(vK.values)
        im = plt.scatter(
            vSAA[idx_sort] * 180 / np.pi, vSEA[idx_sort] * 180 / np.pi,
            s=1, c=vK[idx_sort],
            cmap=COLORMAP_SHADING,
            marker='s', alpha=.5)

        plt.ylabel('Solar elevation angle [°]', fontsize=FONT_SIZE)
        plt.xlabel('Solar azimuth angle [°]', fontsize=FONT_SIZE)

        if self.horizons is not None:
            plt.plot(self.horizons.AZIMUT, self.horizons.ELEVATION, '-', linewidth=1, alpha=0.6, c='red')
            plt.plot(self.horizons.AZIMUT - 360, self.horizons.ELEVATION, '-', linewidth=1, alpha=0.6, c='red')
        if self.latitude < 0:
            dxx = 180
        else:
            dxx = 0

        plt.xlim((45 - dxx, 315 - dxx))
        axe.text(50 - dxx, 0.92 * SELMax, 'shadow analysis')
        im.set_clim(0, cmax)
        plt.ylim((0, SELMax))
        plt.colorbar(im, label=label)

def get_version() :
    # TODO
    #return metadata.metadata('libinsitu')['Version']
    return "1.2"

def _get_meta(df, keys) :
    """Try several keys to get Meta data"""
    for key in keys :
        if key in df.attrs :
            return df.attrs[key]
    return "-"



# -*- coding: utf-8 -*-
"""
Created on Thu Sep  8 12:59:01 2022

@author: y-m.saint-drenan
"""
import os
import pandas as pd
import copy
import numpy as np
from datetime import datetime

from dateutil.relativedelta import relativedelta

from .base_handler import InSituHandler
from libinsitu.log import info, debug
from ..common import GLOBAL_VAR, DIRECT_VAR, parseTimezone

IRRADIANCE_COL="Irradiance(W/m2)"



class SkyNetHandler(InSituHandler) :

    def _read_chunk(self, stream):

        out_col, year = read_header(stream)

        df = pd.read_csv(stream, sep=" ")

        # SKYNET data are every 10 seconds
        # They are offset by 2 seconds to "round" seconds
        # We offset them to round seconds
        df["time"] = pd.to_datetime(dict(
            year=year,
            month=df['Month'],
            day=df['Day']
        )) + pd.to_timedelta((df["Hour"] * 360).round() * 10, unit="seconds")

        df = df.rename(columns={IRRADIANCE_COL: out_col})
        df = df.set_index("time")

        # Apply timezone
        df.index -= parseTimezone(self.properties["Station_Timezone"])

        debug(df.tail())

        return df[[out_col]]


def read_header(stream) :

    out_col = None
    year = None

    # Parse header
    for line in stream:

        line = line.decode()

        # File type
        if "irradiance" in line:
            if "global" in line:
                out_col = GLOBAL_VAR
            elif "direct" in line:
                out_col = DIRECT_VAR

        # Year
        if "Year" in line:
            key, year = line.split(":")
            year = int(year.strip())

        # End of header
        if line.startswith("------"):
            return out_col, year




import argparse
from datetime import datetime

import matplotlib.pyplot as plt
from dateutil.relativedelta import relativedelta
from dotenv import load_dotenv

from libinsitu import openNetCDF, getNetworkId, readShortname, info, LATITUDE_VAR, LONGITUDE_VAR, ELEVATION_VAR
from libinsitu.common import netcdf_to_dataframe
from libinsitu.log import LogContext
from libinsitu.qc.qc_utils import flagData, write_flags, cleanup_data, visual_qc, compute_sun_pos


def parser() :
    parser = argparse.ArgumentParser(description='Perform QC analysis on input file. It can fill QC flags in it and / or generate visual QC image')
    parser.add_argument('input', metavar='<file.nc|odap_url>', type=str, help='Input local file or URL')
    parser.add_argument('--output', '-o', metavar='<out.png>', type=str, help='Output image')
    parser.add_argument('--update', '-u', action="store_true", help='Update QC flags on input file', default=False)
    parser.add_argument('--from-date', '-f', metavar='<yyyy-mm-dd>', type=datetime.fromisoformat, help='Start date on analysis (last 5 years of data by default for graph output)', default=None)
    parser.add_argument('--to-date', '-t', metavar='<yyyy-mm-dd>', type=datetime.fromisoformat, help='End date of analysis', default=None)
    parser.add_argument('--with-mc-clear', '-wmc', action="store_true", help='Enable display of mcClear', default=False)
    parser.add_argument('--with-horizons', '-wh', action="store_true", help='Enable display of horizons', default=False)
    return parser

def main() :

    # Required to load CAMS email
    load_dotenv()

    args = parser().parse_args()

    # Open in read or update mode
    mode = 'a' if args.update else 'r'
    ncfile = openNetCDF(args.input, mode=mode)

    # try to read network and station id from file
    network_id = getNetworkId(ncfile)
    station_id = readShortname(ncfile)

    with LogContext(network=network_id, station_id=station_id, file=args.input) :

        info("Start of QC")

        params = dict(
            start_time=args.from_date,
            end_time=args.to_date,
            rename_cols=True)

        if args.from_date is None and args.output is not None :
            # By default, show 5 years of data in graph ouptput
            params["rel_start_time"] = relativedelta(years=-5)

        # Load NetCDF timeseries as pandas Dataframe
        df = netcdf_to_dataframe(ncfile, **params)

        if args.output :

            visual_qc(
                df,
                with_horizons=args.with_horizons,
                with_mc_clear=args.with_mc_clear)

            # Save to output file
            plt.savefig(args.output)
            plt.close()

        if args.update :

            lat = float(df.attrs[LATITUDE_VAR])
            lon = float(df.attrs[LONGITUDE_VAR])
            alt = float(df.attrs[ELEVATION_VAR])


            # Update NetCDF file with QC
            df = cleanup_data(df)
            sp_df = compute_sun_pos(df, lat, lon, alt)
            flags_df = flagData(df, sp_df)


            write_flags(ncfile, flags_df)







# -*- coding: utf-8 -*-
"""
Created on Thu Jun 23 10:09:54 2022

@author: y-m.saint-drenan
"""
import os
from enum import Enum
from urllib.request import urlopen

import pandas as pd
import pvlib
import sg2
from appdirs import user_cache_dir
from diskcache import Cache
from pandas import DataFrame

from libinsitu import CDL_PATH, read_res, DefaultDict, datetime64_to_sec, seconds_to_idx, getTimeVar, QC_FLAGS_VAR, \
    STATION_ID_ATTRS, STATION_NAME_VAR
from libinsitu.cdl import parse_cdl, initVar
from libinsitu.common import LATITUDE_VAR, LONGITUDE_VAR, ELEVATION_VAR, GLOBAL_VAR, DIFFUSE_VAR, DIRECT_VAR, \
    GLOBAL_TIME_RESOLUTION_ATTR
from libinsitu.log import warning, info
import numpy as np

from libinsitu.qc.matplot import MaplotLibGraphs
from libinsitu.qc.matplot.graphs import _get_meta
from libinsitu.qc.plotly import PlotlyGraphs

cachedir = user_cache_dir("libinsitu")
cache = Cache(cachedir)

MIN_VAL = -100.0
MAX_VAL = 5000.0

CAMS_EMAIL_ENV = "CAMS_EMAIL"


def flagData(meas_df, sp_df):
    """
    :param meas_df: In situ measurements
    :param sp_df: Sun pos / theoretical measurements
    :return: QC flags. -1: no processed. 0: processed and ok. 1: Processed and failed
    """

    MinDailyShareFlag = 0.2

    # Aliases
    GHI = meas_df.GHI
    DIF = meas_df.DHI
    DNI = meas_df.BNI

    TOA = sp_df.TOA
    TOANI = sp_df.TOANI
    GAMMA_S0 = sp_df.GAMMA_S0

    GHI_est = DIF + DNI * np.cos(sp_df.THETA_Z)
    SZA = sp_df.THETA_Z * 180 / np.pi

    size = len(meas_df.GHI)

    KT = np.zeros(size)
    KT[TOA >= 1] = GHI[TOA >= 1] / TOA[TOA >= 1]

    Kn = np.zeros(size)
    Kn[TOANI >= 1] = DNI[TOANI >= 1] / TOANI[TOANI >= 1]

    K = np.zeros(size)
    K[GHI >= 1] = DIF[GHI >= 1] / GHI[GHI >= 1]

    # % % -----------   Calculation of the individual QC flags -----------------
    # BSRN one-component test
    flag_df = DataFrame(index=meas_df.index)
    flag_df["T1C_ppl_GHI"] = (TOA > 0) & (
            (GHI <= -4) | (GHI > 1.5 * TOANI * np.sin(GAMMA_S0) ** 1.2 + 100))
    flag_df["T1C_erl_GHI"] = (TOA > 0) & (
            (GHI <= -2) | (GHI > 1.2 * TOANI * np.sin(GAMMA_S0) ** 1.2 + 50))
    flag_df["T1C_ppl_DIF"] = (TOA > 0) & (
            (DIF <= -4) | (DIF > 0.95 * TOANI * np.sin(GAMMA_S0) ** 1.2 + 50))
    flag_df["T1C_erl_DIF"] = (TOA > 0) & (
            (DIF <= -2) | (DIF > 0.75 * TOANI * np.sin(GAMMA_S0) ** 1.2 + 30))
    flag_df["T1C_ppl_DNI"] = (TOA > 0) & ((DNI <= -4) | (DNI > TOANI))
    flag_df["T1C_erl_DNI"] = (TOA > 0) & (
            (DNI <= -2) | (DNI > 0.95 * TOANI * np.sin(GAMMA_S0) ** 0.2 + 10))

    flag_df["Kn"] = Kn
    flag_df["K"] = K
    flag_df["KT"] = KT

    # BSRN two-component test
    flag_df["T2C_bsrn_kt"] = ((TOA > 0) & (GHI > 50)) & (
                ((SZA < 75) & (K > 1.05)) |
                ((SZA >= 75) & (K > 1.1)))

    # SERI-QC two-component test
    flag_df["T2C_seri_kn_kt"] = (TOA > 0) & ((Kn > KT) | (Kn > 0.8) | (KT > 1.35))
    flag_df["T2C_seri_k_kt"] = (TOA > 0) & (
                ((KT < 0.6) & (K > 1.1)) | ((KT >= 0.6) & (K > 0.95)) | (KT > 1.35))

    # BSRN three-component test
    flag_df["T3C_bsrn_3cmp"] = (TOA > 0) & (
                ((SZA <= 75) & (GHI > 50) & (np.abs(GHI / GHI_est - 1) > 0.08)) | (
                    (SZA > 75) & (GHI > 50) & (np.abs(GHI / GHI_est - 1) > 0.15)))

    # Tracker off test
    GHI_clear = 0.8 * TOA
    DIF_clear = 0.165 * GHI_clear
    DNI_clear = GHI_clear - DIF_clear

    flag_df["tracker_off"] = ((SZA <= 85) &
                                    ((GHI_clear - GHI) / (GHI_clear + GHI) < 0.2) &
                                    ((DNI_clear - DNI) / (DNI_clear + DNI) > 0.95))
    # % % Combination of individual QC tests

    # if at least one of the test is positive, we flag all data (to be eventually refined)
    flag_df["QCtot"] = flag_df["T1C_erl_GHI"] | flag_df["T1C_erl_DIF"] | \
                       flag_df["T1C_erl_DNI"] | flag_df["T2C_bsrn_kt"] | \
                       flag_df["T2C_seri_kn_kt"] | flag_df["T2C_seri_k_kt"] | \
                       flag_df["T3C_bsrn_3cmp"] | flag_df["tracker_off"]

    # Evalue the share of flag data per day
    DailyFlagStat = flag_df["QCtot"].resample('D').sum() / (TOA > 0).resample('D').sum()

    # filter if at least on test fail or the number of flag per day exceeds the minimal share
    flag_df["QCfinal"] = flag_df["QCtot"] | np.in1d(flag_df.index.normalize(),
                                                    DailyFlagStat[DailyFlagStat > MinDailyShareFlag].index.normalize())

    return flag_df


def qc_stats(meas_df, sp_df, flag_df) :

    GHI = meas_df.GHI
    DIF = meas_df.DHI
    DNI = meas_df.BNI

    TOA = sp_df.TOA

    def percent(flags, *components) :
        filt = TOA > 0
        for component in components :
            filt = filt & (component > -2)

        tot = sum(filt)
        if tot == 0 :
            return np.nan
        else:
            return sum(flags & filt) / tot * 100

    return  {
        'T1C_erl_GHI': percent(flag_df.T1C_erl_GHI, GHI),
        'T1C_ppl_GHI': percent(flag_df.T1C_ppl_GHI, GHI),
        'T1C_erl_DIF': percent(flag_df.T1C_erl_DIF, DIF),
        'T1C_ppl_DIF': percent(flag_df.T1C_ppl_DIF, DIF),
        'T1C_erl_DNI': percent(flag_df.T1C_erl_DNI, DNI),
        'T1C_ppl_DNI': percent(flag_df.T1C_ppl_DNI, DNI),
        'T2C_bsrn_kt': percent(flag_df.T2C_bsrn_kt, GHI),
        'T2C_seri_knkt': percent(flag_df.T2C_seri_kn_kt, DNI, GHI),
        'T2C_seri_kkt': percent(flag_df.T2C_seri_k_kt, DIF, GHI),
        'T3C_bsrn': percent(flag_df.T3C_bsrn_3cmp, GHI, DIF, DNI)}


def cleanup_data(df, freq=None):
    """Cleanup and resample data"""

    # Default resolution : take the one from the source
    if freq is None:
        freq = df.attrs[GLOBAL_TIME_RESOLUTION_ATTR]

    # Not UTC ?
    if df.index.tz is not None:
        df.index = df.index.tz_convert('UTC').tz_localize(None)

    # Fill out of range values with NAN
    # XXX use "range" QC check instead
    for varname in [GLOBAL_VAR, DIFFUSE_VAR, DIRECT_VAR] :
        if varname in df :
            var = df[varname]
            df.loc[var > MAX_VAL, varname] = np.nan
            df.loc[var < MIN_VAL, varname] = np.nan
        else:
            warning("Missing var %s, adding NaNs" % varname)
            df[varname] = np.nan

    freq_s = str(freq) + "S"

    df = df.resample(freq_s).ffill()
    df = df.asfreq(freq_s)

    start_date = df.index.min().normalize()
    end_date = df.index.max().normalize() + np.timedelta64(24 * 60 - 1, "m")

    df = df.reindex(pd.date_range(start_date, end_date, freq=freq_s))

    return df

#@cache.memoize()
def sun_position(lat, lon, alt, start_time, end_time, freq="60S") :

    if alt == np.nan:
        alt = 0

    times = pd.date_range(start_time, end_time, freq=freq)

    sun_rise = sg2.sun_rise(
        [[lon, lat, alt]],
        times)

    sun_pos = sg2.sun_position(
        [[lon, lat, alt]],
        times,
        ["topoc.alpha_S", "topoc.gamma_S0", "topoc.toa_hi", "topoc.toa_ni"])

    SR = np.squeeze(sun_rise[:, 0, 0])
    SR_Day = SR.astype('datetime64[D]').astype(SR.dtype)
    SR_TOD = (SR - SR_Day).astype(float) / 1000 / 60 / 60

    SS = np.squeeze(sun_rise[:, 0, 2])
    SS_Day = SS.astype('datetime64[D]').astype(SS.dtype)
    SS_TOD = (SS - SS_Day).astype(float) / 1000 / 60 / 60

    df = pd.DataFrame(index=times)

    # Add extra columns from SG2 to dataframe
    df['THETA_Z'] = np.pi / 2 - np.squeeze(sun_pos.topoc.gamma_S0)
    df['GAMMA_S0'] = np.squeeze(sun_pos.topoc.gamma_S0)
    df['ALPHA_S'] = np.squeeze(sun_pos.topoc.alpha_S)
    df['SZA'] = 90 - 180 / np.pi * np.squeeze(sun_pos.topoc.gamma_S0)
    df['TOA'] = np.squeeze(sun_pos.topoc.toa_hi)
    df['TOANI'] = np.squeeze(sun_pos.topoc.toa_ni)
    df['SR_h'] = SR_TOD
    df['SS_h'] = SS_TOD

    return df


@cache.memoize()
def get_cams(start_date, end_date, lat, lon, altitude, time_step="1min") :

    info("Calling CAMS")

    if CAMS_EMAIL_ENV in os.environ:
        cams_email = os.environ[CAMS_EMAIL_ENV]
    else:
        raise Exception("Cams emails not found. Please set the env variable %s or use a .env file" % CAMS_EMAIL_ENV)

    CAMS_DF, _ =  pvlib.iotools.get_cams(
                start=start_date,
                end=end_date,
                latitude=lat, longitude=lon,
                email=cams_email,
                identifier='mcclear',
                altitude=altitude, time_step=time_step, time_ref='UT', verbose=False,
                integrated=False, label='right', map_variables=True,
                server='www.soda-is.com', timeout=180)

    res = pd.DataFrame({
        'CLEAR_SKY_GHI': CAMS_DF.ghi_clear.values,
        'CLEAR_SKY_DNI': CAMS_DF.dni_clear.values,
        'CLEAR_SKY_DIF': CAMS_DF.dhi_clear.values},
        index=CAMS_DF.index.values)

    info("End calling CAMS")

    return res

@cache.memoize()
def wps_Horizon_SRTM(lat, lon, altitude):
    if np.abs(lat) < 60 :
        return None

    info("Fetching horizons from WPS")

    str_wps = 'http://toolbox.webservice-energy.org/service/wps?service=WPS&request=Execute&identifier=compute_horizon_srtm&version=1.0.0&DataInputs='
    datainputs_wps = 'latitude={:.6f};longitude={:.6f};altitude={:.1f}'.format(lat, lon, altitude)

    response = urlopen('{}{}'.format(str_wps, datainputs_wps))

    HZ = pd.read_csv(response, delimiter=';', comment='#', header=None, skiprows=17, nrows=360,
                         names=['AZIMUT', 'ELEVATION'])

    info("Horizons fetched")

    return HZ



def write_flags(ncfile, flags_df) :

    # Parse CDL : use defaultdict to avoid warning
    # XXX try to not parse it twice and get it from above
    cdl = parse_cdl(read_res(CDL_PATH), attributes=DefaultDict(lambda : "-"))

    # Create var if not present yet
    if not QC_FLAGS_VAR in ncfile.variables :
        initVar(ncfile, cdl.variables[QC_FLAGS_VAR])

    qc_var = ncfile.variables[QC_FLAGS_VAR]

    # Build a dictionary of masks
    flag_masks = dict((flag, mask) for flag, mask in zip(qc_var.flag_meanings.split(), qc_var.flag_masks))

    info("Flag masks  : %s" % flag_masks)

    # Output
    out_masks = np.zeros(len(flags_df))

    for colname in flags_df.columns :
        if not colname in flag_masks :
            warning("Flag %s not found in QC flags DSL. Skipping" % colname)
            continue

        colvalues = flags_df[colname]

        out_masks += colvalues.values * flag_masks[colname]

    # Compute IDX
    dates = flags_df.index.values
    times_sec = datetime64_to_sec(ncfile, dates)
    time_idx = seconds_to_idx(ncfile, times_sec)

    # Assign flags
    time_var = getTimeVar(ncfile)
    max_time = len(time_var)

    out_idx = time_idx > max_time -1
    if np.any(out_idx) :
        warning("Index of of time range. Truncating %d values" % np.sum(out_idx))
        time_idx = time_idx[~out_idx]
        out_masks = out_masks[~out_idx]

    qc_var[time_idx] = out_masks

def compute_sun_pos(df, lat, lon, alt) :
    """Call sg2 on data"""

    # Compute geom & theoretical irradiance
    sp_df = sun_position(
        lat, lon, alt,
        df.index.min(),
        df.index.max(),
        freq=pd.infer_freq(df.index))

    return sp_df

class ShowFlag(Enum):
    HIDE="hide" # Hide data with errors
    SHOW="show" # Show all data
    FLAG="flag" # Flag errors in red

def visual_qc(
        df,
        latitude = None,
        longitude = None,
        elevation = None,
        station_id = None,
        station_name = None,
        with_horizons = False,
        with_mc_clear = False,
        show_flag=ShowFlag.SHOW,
        engine="matplotlib"):

    """
    Generates matplotlib graphs for visual QC

    :param df: Dataframe of input irradiance. It should have a time index and 3 columns : GHI, DHI, BNI).
               This dataframe can typically be obtained with netcdf_to_dataframe(... rename_cols=True)
    :param latitude: Latitude of the station. Can also be passed as meta data (.attrs) of the Dataframe
    :param longitude: Longitude of the station. Can also be passed as meta data (.attrs) of the Dataframe
    :param elevation: elevation of the station. Can also be passed as meta data (.attrs) of the Dataframe
    :param station_id: Id of the station (optional). Can also be passed as meta data (.attrs) of the Dataframe
    :param station_id: Name of the station (optional). Can also be passed as meta data (.attrs) of the Dataframe
    :param with_horizons: True to compute horizons (requires network)
    :param with_mc_clear: True to compute mc_clear from SODA (requires SODA credentials and network access)
    """

    # Resample to the minute to produce graph
    resolution_sec = 60

    # Clean data
    df = cleanup_data(df, resolution_sec)

    # Get meta data from parameters or from attributes attached to the Dataframe
    lat = latitude if latitude else  float(df.attrs[LATITUDE_VAR])
    lon = longitude if longitude else float(df.attrs[LONGITUDE_VAR])
    alt = elevation if elevation else float(df.attrs[ELEVATION_VAR])
    station_id = station_id if station_id else _get_meta(df, STATION_ID_ATTRS)
    station_name = station_name if station_name else _get_meta(df, STATION_NAME_VAR)

    # Compute geom & theoretical irradiance
    sp_df = compute_sun_pos(df, lat , lon, alt)

    # Compute QC flags
    flags_df = flagData(df, sp_df)

    # Fetch horizons
    if with_horizons:
        horizons = wps_Horizon_SRTM(lat, lon, alt)
    else:
        horizons = None

    if with_mc_clear:
        cams_df = get_cams(
            start_date=df.index.min(),
            end_date=df.index.max(),
            lat=lat, lon=lon,
            altitude=alt)
        cams_df = cams_df.reindex(df.index)
    else:
        cams_df = None

    # Transform flag
    # TODO : refactor main_layout to use Enum too
    flag = {
        ShowFlag.SHOW : 0,
        ShowFlag.HIDE : -1,
        ShowFlag.FLAG : 1
    }[show_flag]

    # Statistics on QC flags
    stat_test = qc_stats(df, sp_df, flags_df)

    # Pick class depending on engine
    Clazz = {
        "matplotlib" : MaplotLibGraphs,
        "plotly" : PlotlyGraphs
    }[engine]

    # Draw figures
    graph = Clazz(
        meas_df=df,
        sp_df=sp_df,
        flag_df=flags_df,
        cams_df=cams_df,
        horizons=horizons,
        stat_test=stat_test,
        latitude=lat, longitude=lon, elevation=alt,
        station_id=station_id, station_name=station_name,
        show_flag=flag)

    return graph.main_layout()




# -*- coding: utf-8 -*-
"""
Created on Fri Sep  9 11:00:43 2022

@author: y-m.saint-drenan
"""
ListStations={
'94002':{'Name':'Portland DEQ, OR','abbrev':'P1'},
'94003':{'Name':'Milwaukie MES, OR','abbrev':'PL'},
'94005':{'Name':'Gladstone, OR','abbrev':'GL'},
'94007':{'Name':'Scoggins Creek, OR','abbrev':'SC'},
'94008':{'Name':'Forest Grove, OR','abbrev':'FG'},
'94019':{'Name':'Aprovecho, OR','abbrev':'AP'},
'94040':{'Name':'Ashland, OR','abbrev':'AS'},
'94101':{'Name':'Green River, WY','abbrev':'GR'},
'94102':{'Name':'Moab, UT','abbrev':'MO'},
'94145':{'Name':'Dillon, MT','abbrev':'DI'},
'94158':{'Name':'Cheney, WA','abbrev':'CY'},
'94166':{'Name':'Klamath Falls, OR','abbrev':'KF'},
'94167':{'Name':'Whitehorse Ranch, OR','abbrev':'WH'},
'94168':{'Name':'La Grande, OR','abbrev':'LG'},
'94169':{'Name':'Hermiston, OR','abbrev':'HE'},
'94170':{'Name':'Burns, OR','abbrev':'BU'},
'94171':{'Name':'Twin Falls (Kimberly), ID','abbrev':'TF'},
'94172':{'Name':'Picabo, ID','abbrev':'PI'},
'94173':{'Name':'Parma, ID','abbrev':'PA'},
'94174':{'Name':'Aberdeen, ID','abbrev':'AB'},
'94181':{'Name':'Coeur d Alene, ID','abbrev':'CD'},
'94182':{'Name':'Boise, ID','abbrev':'BO'},
'94185':{'Name':'Challis, ID','abbrev':'BO'},
'94249':{'Name':'Silver Lake, OR','abbrev':'SL'},
'94250':{'Name':'Klamath Falls, OR','abbrev':'KF'},
'94251':{'Name':'Christmas Valley, OR','abbrev':'CH'},
'94252':{'Name':'Madras, OR','abbrev':'MA'},
'94253':{'Name':'Corvallis, OR','abbrev':'CV'},
'94254':{'Name':'Willamette High School, Eugene, OR','abbrev':'WI'},
'94255':{'Name':'Eugene, OR','abbrev':'EU'},
'94256':{'Name':'Bend, OR','abbrev':'BE'},
'94257':{'Name':'Coos Bay, OR','abbrev':'CB'},
'94258':{'Name':'Portland, OR','abbrev':'PT'},
'94277':{'Name':'Hood River, OR','abbrev':'HR'},
'94278':{'Name':'West Hood River, OR','abbrev':'WR'},
'94279':{'Name':'Parkdale, OR','abbrev':'PD'}}

import os
import pandas as pd
import numpy as np
import copy

pathIN='V:\\IN_SITU_data\\RawData\\UniOregon\\'
pathOUT='.//'

# List the available stations
listFiles=os.listdir(pathIN)
listStat=list(set([xx[17:-11] for xx in listFiles]))

# rank of the station to read
iStation=0

# list the data available for the current station
listFiles=os.listdir(pathIN)
listFiles=[xx for xx in listFiles if listStat[iStation] in xx]

# for-loop on the different data corresponding to the station
for iFile,yyy in enumerate(listFiles):
    
    #read the archive
    df=pd.read_csv(pathIN+listFiles[iFile],sep='\t')
    col=df.columns
    StatNr=col[0]
    df_year=col[1]
    df=df.rename(columns = {col[0]:'doy',col[1]:'MilitaryTime'})
    col2=df.columns
    col2=[xx for xx in col2 if xx[0]!='0' ]
    df=df[col2]
    
    # decode the time format
    df_hours=(np.floor(df['MilitaryTime']/100)).astype(int)
    df_minutes=df['MilitaryTime']-df_hours*100
    doy_sec=df['doy'].values*24*60*60+df_hours*60*60+df_minutes*60
    df['time']=pd.to_datetime([np.datetime64(str(col[1])+'-01-01')+ np.timedelta64(x,'s') for x in doy_sec])
    df=df.set_index('time')
    df=df.drop(columns=['doy','MilitaryTime'])
    
    # Decode the name of the columns
    #http://solardat.uoregon.edu/DataElementNumbers.html
    col=df.columns
    DictIDs={'GHI':'100','DNI':'201','DIF':'300','PRES':'917',\
             'WD':'920','WS':'921','Ta':'930','Td':'931','RH':'933'}
    for prm in DictIDs:
        cnt=0
        for ii in range(10):
            ID='{}{}'.format(DictIDs[prm],ii)
            if ID in col:
                if cnt==0:
                    df=df.rename(columns = {ID:prm})
                else:
                    df=df.rename(columns = {ID:prm+'_'+str(cnt)})
                cnt+=1
                        
    
    # concatenate the extracted data
    if iFile==0:
        df_UniOregon=copy.deepcopy(df)
    else:
        df_UniOregon=pd.concat([df_UniOregon,df])


# compile the metadata
metadata={'StationID':listStat[iStation],'StationNr':StatNr,'Name':ListStations[str(StatNr)]['Name'],'Abbrev':ListStations[str(StatNr)]['abbrev']}

# prepare xarray with metadata and write the netcdf
xr_UniOregon=df_UniOregon.to_xarray()
xr_UniOregon.attrs['Network']='University of Oregon'
for xx in metadata:
    xr_UniOregon.attrs[xx]=metadata[xx]
xr_UniOregon.to_netcdf(pathOUT+"UniOregon_{}.nc".format(metadata['StationID']))



import os, sys
from collections import defaultdict

from dateutil.relativedelta import relativedelta
import argparse
import numpy as np
from numpy import datetime64
from pandas import Series
from rich.console import Console
from rich.table import Table
from datetime import datetime

from libinsitu import qc_masks
from libinsitu.log import debug
from libinsitu.common import netcdf_to_dataframe, CHUNK_SIZE, df_to_csv, QC_FLAGS_VAR

QC_NONE = "none"
QC_MASK = "masks"
QC_NAMES = "names"
QC_EXPAND = "expand"

DATE_FORMATS_PARTS = [
    ("%Y", 4, "years"),
    ("-%m", 3, "months"),
    ("-%d", 3, "days"),
    ("T%H", 3, "hours"),
    (":%M", 3, "seconds")]

DATE_FORMATS = dict()
full_size=0
full_format=""
for part, size, name in DATE_FORMATS_PARTS :
    full_size += size
    full_format += part
    DATE_FORMATS[full_size] = (full_format, name)


def parse_date_filter(strval) -> (datetime64, datetime64):
    """Parse partial date to from/to datetimes"""
    length = len(strval)
    if not length in DATE_FORMATS :
        raise Exception('Invalid date time filter. THe following format are supported : %s' % ', '.join(format for format, name in DATE_FORMATS.values()))
    format, name = DATE_FORMATS[length]

    start = datetime.strptime(strval, format)
    end = start + relativedelta(**{name:1})

    debug(format, name, start, end)

    return np.datetime64(start), np.datetime64(end)

class Stat() :
    def __init__(self):
        self.min = np.nan
        self.max = np.nan
        self.sum = 0.0
        self.count = 0

    def accumulate(self, series):
        minval = series.min()
        if minval is not np.nan :
            self.min = np.nanmin([self.min, minval])

        maxval = series.max()
        if maxval is not np.nan:
            self.max = np.nanmax([self.max, maxval])


        self.sum += series.sum()
        self.count += series.count()

def print_meta(df) :

    def print_dict(dic, indent=0) :
        for key, val in dic.items() :
            if isinstance(val, dict) :
                print("#%s %s:" % (" " * indent, key))
                print_dict(val, indent+2)
            else:
                print("#%s %s = %s" % (" " * indent, key, str(val)))

    print_dict(df.attrs)


def float_to_str(nb_digits) :
    format = "%." + str(nb_digits) + "f"

    def format_f(floatval) :
        return format % floatval

    return format_f

def build_formatters(df) :
    res = dict()
    for varname in df.columns :
        if not varname in df.attrs["variables"] :
            continue
        var_attrs = df.attrs["variables"][varname]
        if "least_significant_digit" in var_attrs :
            res[varname] = float_to_str(var_attrs["least_significant_digit"])
    return res


def parser() :

    parser = argparse.ArgumentParser(description='Dump content of NetCDF insitu data (CF compliant)')
    parser.add_argument('filename', metavar='<file.nc> or <http://opendap-url/.nc>', type=str, help='Input file or URL')
    parser.add_argument('--type', '-t', choices=["csv", "text"], help='Output type', default="text")
    parser.add_argument('--skip-na', '-s', action='store_true', help="Skip lines with only NA values", default=False)
    parser.add_argument('--skip-qc', '-sq', action='store_true', help="Skip lines bad QC", default=False)
    parser.add_argument('--filter', '-f', metavar="'<time> or <from_time>~<to-time>, with any sub part of 'YYYY-mm-ddTHH:MM:SS'", help="Time filter")
    parser.add_argument('--qc-format', '-qf', metavar="Format for QC flags (none, mask, names or expand)", choices=[QC_NONE, QC_MASK, QC_NAMES, QC_EXPAND], help="Format for QC flags", default=None)
    parser.add_argument('--stats', '-z', action="store_true", default=False, help="Performs statistics. Don't print data")
    parser.add_argument('--header', '-hd', action="store_true", default=False, help="Dump global and var meta data as header")
    parser.add_argument('--no-data', '-n', action="store_true", default=False, help="Don't print data. Useless together with --header to print meta data only")
    parser.add_argument('--cols', '-c', metavar="<col1>,<col2> ..", help="Selection of columns. All by default")
    parser.add_argument('--user', '-u', help='User login (or TDS_USER env var), for URL',
                        default=os.environ.get("TDS_USER", None))
    parser.add_argument('--password', '-p', help='User password (or TDS_PASS env var), for URL',
                        default=os.environ.get("TDS_PASS", None))
    parser.add_argument('--steps', '-st', help='Downsampling', type=int, default=1)
    parser.add_argument('--chunk_size', '-cs', help='Size of chunks', type=int, default=CHUNK_SIZE)
    return parser

def main() :

    args = parser().parse_args()
    cols = args.cols.split(",") if args.cols else None

    fromTime=None
    toTime=None

    if args.filter :
        if "~" in args.filter :
            filter1, filter2 = args.filter.split("~")
            fromTime, _ = parse_date_filter(filter1)
            _, toTime = parse_date_filter(filter2)
        else :
            fromTime, toTime = parse_date_filter(args.filter)

    if args.qc_format is None :
        if args.type == "csv" :
            args.qc_format = QC_EXPAND
        else :
            args.qc_format = QC_NAMES

    chunks = netcdf_to_dataframe(
        args.filename,
        fromTime, toTime,
        user=args.user, password=args.password,
        drop_duplicates=True,
        skip_na=args.skip_na,
        skip_qc=args.skip_qc,
        vars=cols,
        chunked=True,
        steps=args.steps, chunk_size=args.chunk_size)

    if args.stats :

        show_stats(chunks)

    else:

        print_data(chunks, args)

def print_data(chunks, args) :
    header = True
    formatters = None
    for chunk in chunks:

        if len(chunk) == 0:
            continue

        chunk = format_QC(chunk, args.qc_format)

        if header and args.header:
            print_meta(chunk)

        if formatters is None:
            formatters = build_formatters(chunk)

        if args.no_data:
            break

        if args.type == "text" :
            chunk.to_string(sys.stdout, justify="left", header=header, formatters=formatters)
            print("")
        elif args.type == "csv" :
            df_to_csv(chunk, index_label="time", header=header)

        header = False

def show_stats(chunks) :

    console = Console()

    stats = defaultdict(lambda : Stat())
    for chunk in chunks :

        chunk = format_QC(chunk, QC_EXPAND)

        for col in chunk.columns :
            series = chunk[col]
            stats[col].accumulate(series)

    # Data vars stats
    table = Table("column", "count", "min", "max", "mean")

    has_qc = False

    for colname, stat in stats.items() :

        if colname.startswith(QC_FLAGS_VAR) :
            has_qc = True
            continue

        table.add_row(
            colname,
            "%d" % stat.count,
            "%.05g" % stat.min,
            "%.05g" % stat.max,
            "%.05g" % (stat.sum / stat.count))

    console.print(table)

    if has_qc :

        table = Table("QC flag", "fail", "%")
        for colname, stat in stats.items() :

            if not colname.startswith(QC_FLAGS_VAR) :
                continue

            table.add_row(
                colname.strip(QC_FLAGS_VAR + "."),
                "%d" % stat.sum,
                "%.02f" % (stat.sum / stat.count * 100))

        console.print(table)

    # QC stats



def format_QC(df, qc_format) :

    if not QC_FLAGS_VAR in df.columns :
        return df

    qc_col = df[QC_FLAGS_VAR]

    masks_dict = qc_masks(df)

    if qc_format == QC_MASK :
        res = Series(data="", index = df.index, dtype=str)
        col_names = []
        for idx, (flag, mask) in enumerate(masks_dict.items()) :
            letter = chr(97+idx)
            col_names.append("%s:%s" % (flag, letter))
            res += np.where(qc_col.values & mask != 0, letter, ".")

        df[QC_FLAGS_VAR] = res

        # Rename column to provide details
        col_name = "%s[%s]" % (QC_FLAGS_VAR, ";".join(col_names))
        df = df.rename(columns={QC_FLAGS_VAR: col_name})

    elif qc_format == QC_EXPAND :

        for flag, mask in masks_dict.items():
            colname = "%s.%s" % (QC_FLAGS_VAR, flag)
            df[colname] = np.where(qc_col & mask == 0, 0, 1)
        del df[QC_FLAGS_VAR]

    elif qc_format == QC_NAMES :
        res = Series(data="", index=df.index, dtype=np.object)
        for flag, mask in masks_dict.items():
            res += np.where(
                qc_col & mask == 0,
                "",
                np.where(
                    res == "",
                    flag ,
                    ";" + flag))
        df[QC_FLAGS_VAR] = res

    elif qc_format == QC_NONE :
        del df[QC_FLAGS_VAR]

    else:
        raise Exception("Unkown QC format : %s" % qc_format)

    return df

if __name__ == '__main__':
    main()


# Performs various checks on NetCDF files
import argparse
from netCDF4 import Dataset

from libinsitu import getStationId, getNetworkId
from libinsitu.common import netcdf_to_dataframe, GLOBAL_VAR, DIFFUSE_VAR, DIRECT_VAR, PRESSURE_VAR, \
    HUMIDITY_VAR, TEMP_VAR
from libinsitu.log import *
import numpy as np

NAN_VALUES = {
    GLOBAL_VAR : -999.0,
    DIFFUSE_VAR : -999.0,
    DIRECT_VAR : -999.0,
    TEMP_VAR : 173.25,
    HUMIDITY_VAR : -0.999,
    PRESSURE_VAR: -99900.0
}

CHUNK_SIZE = 100

def file2df(filename) :
    nc = Dataset(filename, mode='r')
    df = netcdf_to_dataframe(nc)
    nc.close()
    return df


def diff(df1, df2, args) :

    identical = True

    if len(df1.index) != len(df2.index) or not np.all(df1.index == df2.index) :
        identical = False
        warning("Timing differ : [%s|%s|%d] <-> [%s|%s|%d]",
                df1.index.min(), df1.index.max(), len(df1.index),
                df2.index.min(), df2.index.max(), len(df2.index))

    join = df1.join(df2, lsuffix="1", rsuffix="2", how="inner")

    cols1 = set(df1.columns)
    cols2 = set(df2.columns)

    if cols1 != cols2 :
        warning("datasets have different set of colums : %s <-> %s" % (",".join(cols1), ",".join(cols2)))

    common_cols = cols1 & cols2

    for col in common_cols :
        with LogContext(file=col) :

            data1 = join[col + "1"]
            data2 = join[col + "2"]

            # Check missing values
            for this, other, this_name, other_name in [(data1, data2, "data1", "data2"), (data2, data1, "data2", "data1")] :

                missing_idx = this.isna() & ~other.isna()
                missing_nb = sum(missing_idx)
                if missing_nb > 0:
                    missing_vals = other[missing_idx]
                    min_missing = min(missing_vals)
                    max_missing = max(missing_vals)

                    identical = False

                    missing_dates = join.index[missing_idx]

                    warning("%d NA values in %s only. Data values in %s : [%f:%f] from %s to %s",
                            missing_nb,
                            this_name,
                            other_name,
                            min_missing,
                            max_missing,
                            np.min(missing_dates),
                            np.max(missing_dates))


            # Check different values
            rmse = np.sqrt(np.mean((data1 - data2) ** 2))
            if rmse > 0 :
                identical = False
                warning("rmse:%f", rmse)

                nonna = ~data1.isna() & ~data2.isna()
                non_equal = nonna & ~np.isclose(data1, data2)
                diff_nb = sum(non_equal)

                if diff_nb > 0 :
                    identical = False

                    dates = join.index[non_equal]

                    warning("Different values in data1 and data2 : %d. From %s to %s",
                            diff_nb,
                            np.min(dates),
                            np.max(dates))

    if identical :
        info("The two datasets are identical")

def dump(df1, df2) :

    debug("Join")
    outer = df1.join(df2, lsuffix="1", rsuffix="2", how="outer")
    mask = np.zeros(outer.index.values.shape)

    for col in df1. columns :
        mask = mask | (outer[col+"1"] != outer[col+"2"])

    debug("Diff")
    diff = outer[mask]

    debug("Print")
    for i in range(0, diff.shape[0], CHUNK_SIZE) :
        chunk = diff[i:i+CHUNK_SIZE]
        chunk.to_string(sys.stdout, header=True, justify="left")
        print("\n")

def main() :

    parser = argparse.ArgumentParser(description='Transform In-Situ data into NetCDF files')
    parser.add_argument('file1', metavar='<file1.nc>', type=str, help='First NetCDF file')
    parser.add_argument('file2', metavar='<file2.nc>', type=str, help='Second NetCDF file')
    parser.add_argument('--dump', '-d', default=False, action='store_true',
                        help="Dump all diff in text format")
    args = parser.parse_args()

    df1 = file2df(args.file1)
    df2 = file2df(args.file2)

    # Replace nan values for df2
    for col, na_val in NAN_VALUES.items():
        if col not in df2 :
            continue
        vals = df2[col]
        idx = np.isclose(vals, na_val)
        df2[col][idx] = np.nan

    station_id = getStationId(df1.attrs)
    network = getNetworkId(df1.attrs)

    with LogContext(network=network, station_id=station_id, file="%s:%s" % (args.file1, args.file2)):

        if args.dump :
            dump(df1, df2)
        else:
            diff(df1, df2, args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
from libinsitu.handlers import HANDLERS
from libinsitu.common import *
from libinsitu.cdl import init_nc
from libinsitu.log import *
import argparse


def update_times(outds, start_date64, dry_run=False) :
    time_var = getTimeVar(outds)
    resolution_s = getTimeResolution(outds)
    start_sec = datetime64_to_sec(outds, start_date64)

    if time_var[0] != start_sec :
        if dry_run :
            info("Would have updated times values. Firs value : %d => %d" % (time_var[0], start_sec))
        else:
            info("Updating time values. Firs value : %d => %d" % (time_var[0], start_sec))
            time_values = np.arange(start_sec, start_sec + resolution_s * len(time_var), resolution_s)
            time_var[:] = time_values
    else:
        info("First time value was corect : no update required")


def update_meta(input, output, network=None, dry_run=False, delete=False, update_time=False) :

    if output is None :
        output = input

    # Output
    new = not os.path.exists(output)

    if dry_run :
        mode= 'r'
    elif new :
        mode = "w"
    else:
        mode = "a"

    outds = Dataset(output, mode=mode)

    # Input
    if input == output :
        inds = outds
    else:
        inds = Dataset(input, mode="r")

    # Guess network
    if network is None :
        network = getattr(inds, "network_id", None)
        if network is None :
            raise Exception("Unable to guess network name. Please provide it")


    station_id = read_str(inds.variables[STATION_NAME_VAR])
    properties = getProperties(network, station_id)

    timeVar = getTimeVar(inds)
    first_time = sec_to_datetime64(inds, timeVar[0]).data[()]
    last_time = sec_to_datetime64(inds, timeVar[-1]).data[()]

    # Attribute with null values are ignored : previous value is kept
    properties["FirstData"] = time2str(first_time, seconds=True)
    properties["LastData"] = time2str(last_time, seconds=True)
    properties["UpdateTime"] = None
    properties["CreationTime"] = None

    handler = HANDLERS[network](properties)

    init_nc(outds, properties, handler.data_vars(), dry_run, delete)

    if inds != outds :

        # copy all file data except for the excluded
        for vname, variable in inds.variables.items():
            if vname in outds.variables and "time" in variable.dimensions :
                info("Copying data of var '%s' from '%s' to '%s'" % (vname, input, output))

                outds.variables[vname][:] = variable[:]

    if update_time :
        start_date = str_to_date64(properties["Station_StartDate"])
        update_times(outds, start_date, dry_run)

def main() :

    parser = argparse.ArgumentParser(description='Update meta attributes in NetCDF file')
    parser.add_argument('input', metavar='<input.nc>', type=str, help='NetCDF file to update')
    parser.add_argument('--output', '-o', metavar='<output.nc>', type=str, help='NetCDF files to update', default=None)
    parser.add_argument('--network', metavar='<network_id>', type=str, help='Network. Guessed from reading the file if not provided')
    parser.add_argument('--dry-run', '-n', help='Do not update anything. Just look what would be done', action='store_true', default=False)
    parser.add_argument('--update-ranges', '-ur', help='Update data time ranges for each variable', action='store_true', default=False)
    parser.add_argument('--delete', '-d', help='Delete extra attributes', action='store_true', default=False)
    args = parser.parse_args()

    update_meta(args.input, args.output, args.network, args.dry_run, args.delete, args.update_ranges)

if __name__ == '__main__':
    main()





import pandas as pd

from pandas import DataFrame

from libinsitu.common import GLOBAL_VAR
from libinsitu.handlers.base_handler import InSituHandler, map_cols

MAPPING = dict(Gg_pyr=GLOBAL_VAR)

class ISEPVLive(InSituHandler) :

    def _read_chunk(self, stream) -> DataFrame:
        data = pd.read_csv(stream,sep='\t',parse_dates=['datetime'],index_col=0)
        data = map_cols(data, MAPPING)
        return data



from plotly.subplots import make_subplots
import plotly.express as px
from libinsitu.qc.matplot import MaplotLibGraphs
import datashader as ds
import datashader.transfer_functions as tf

class PlotlyGraphs(MaplotLibGraphs) :

    def main_layout(self):

        fig = make_subplots(rows=3, cols=1, shared_xaxes=True)

        fig.add_trace(
            self.plot_timeseries("GHI", self.GHI, 1400),
            row=1, col=1)

        fig.add_trace(
            self.plot_timeseries("DNI", self.DNI, 1400),
            row=2, col=1)

        fig.add_trace(
            self.plot_timeseries("DIF", self.DIF, 1000),
            row=3, col=1)

        return fig


    def plot_timeseries(self, label, data, ymax):

        df = data.to_frame("y")
        # Time as int
        df["x"] = data.index.astype('int64')

        x_range = (df.x[0], df.x[-1])
        y_range = (0, ymax)

        cvs = ds.Canvas(x_range=x_range, y_range=y_range, plot_height=100, plot_width=900)

        line = cvs.line(df, 'x', 'y')
        img = tf.shade(line).to_pil()

        print(type(cvs), type(img), cvs, img)

        im = px.imshow(img)

        return im.data[0]



from libinsitu import NA_VALUES
from libinsitu.handlers.base_handler import InSituHandler
import pandas as pd
from libinsitu.log import info, debug

DATE_COLS = dict(
    year='Year',
    month='Month',
    day='Day',
    hour='Hour',
    minute='Minute')

class IEA_PVPSHandler(InSituHandler) :

    def _read_chunk(self, stream):
        df = pd.read_csv(stream, comment="#", na_values=NA_VALUES)

        df["time"] = pd.to_datetime(dict((key, df[val]) for key, val in DATE_COLS.items()))
        df = df.drop(columns=list(DATE_COLS.values()))
        df = df.set_index("time")

        debug(df)

        return df




from collections import OrderedDict
from datetime import timedelta

import pandas as pd

from libinsitu.common import GLOBAL_VAR, DIRECT_VAR, DIFFUSE_VAR, TEMP_VAR, HUMIDITY_VAR, PRESSURE_VAR, WIND_SPEED_VAR, \
    WIND_DIRECTION_VAR, NA_VALUES, parseTimezone
from libinsitu.handlers.base_handler import InSituHandler, ZERO_DEG_K
from libinsitu.log import info, warning


# Possible mapping
MAPPING = dict(
            ghi=GLOBAL_VAR,
            dni=DIRECT_VAR,
            dhi=DIFFUSE_VAR,
            GHI=GLOBAL_VAR,
            DNI=DIRECT_VAR,
            DHI=DIFFUSE_VAR,
            t_air=TEMP_VAR,
            rh=HUMIDITY_VAR,
            bp=PRESSURE_VAR,
            ws=WIND_SPEED_VAR,
            wd=WIND_DIRECTION_VAR)

def read_mesor(stream, na_values=NA_VALUES):

    metadata = {}  # Initilize dictionary containing metadata
    channels = OrderedDict()
    metadata["channels"] = channels


    FirstLine = str(stream.readline())

    if not 'MESOR' in FirstLine:
        raise Exception("This does not appear to be a MESOR file")

    line = ''

    while not ("#begindata" in line) :
        line = stream.readline()

        if ('#comment' in line) or  ('#begindata' in line) :
            continue

        line = line.strip("#")

        if line.startswith("channel") :
            parts = line.split(maxsplit=2)
            channels[parts[1].strip()] = parts[2].strip()
        else:
            parts = line.split(maxsplit=1)
            if len(parts) >= 2 :
                metadata[parts[0].strip()] = parts[1].strip()

    def safe_date(v) :
        return pd.to_datetime(v, errors="coerce")

    # Read data as CSV
    data = pd.read_csv(
        stream,
        header=None,
        delimiter='\t',
        comment='#',
        parse_dates=[0],
        index_col=0,
        na_values=na_values,
        date_parser=safe_date)

    # Bad dates ?
    nb_baddates = sum(data.index.isnull())
    if nb_baddates > 0 :
        warning("Skipping %d bad dates" % nb_baddates)
        data = data[data.index.notnull()]

    data.columns = list(channels.keys())[2:]

    return metadata, data

class EnerMENAHandler(InSituHandler) :

    def _read_chunk(self, stream):

        metadata, data = read_mesor(stream)

        keys = list(key for key in data.columns if key in MAPPING)
        mapping = {key: MAPPING[key] for key in keys}

        data = data[keys]
        data = data.rename(columns=mapping)

        tz = timedelta(hours=0)
        if 'timezone' in metadata :

            tz = parseTimezone(metadata['timezone'])

        if tz != timedelta(hours=0) :
            info("Applying timezone : %s", tz)
            data.index = data.index - tz

        # Convertions
        if TEMP_VAR in data :
            data[TEMP_VAR] = data[TEMP_VAR] + ZERO_DEG_K  # T2: °C -> K
        if HUMIDITY_VAR in data:
            data[HUMIDITY_VAR] = data[HUMIDITY_VAR] / 100  # percent -> 1
        if PRESSURE_VAR in data:
            data[PRESSURE_VAR] = data[PRESSURE_VAR] * 100  # Pressure hPa->Pa

        return data

    def data_vars(self):
        """ @override """
        return [GLOBAL_VAR, DIFFUSE_VAR, DIRECT_VAR, TEMP_VAR, HUMIDITY_VAR, PRESSURE_VAR, WIND_SPEED_VAR, WIND_DIRECTION_VAR]


#!/usr/bin/env python
import os.path
import shutil
import sys
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from tempfile import NamedTemporaryFile
from urllib.request import urlretrieve
import gzip
from dateutil.relativedelta import relativedelta

from libinsitu.common import getStationsInfo, DATE_FORMAT, parse_value, getNetworksInfo, parse_bool
from datetime import datetime, timedelta

from libinsitu.log import info, LogContext, IgnoreAndLogExceptions
import argparse

SOURCE_URL_ATTR="SourceURL"
RAW_PATH_ATTR="RawDataPath"
COMPRESS_ATTR="Compress"

ERROR_SUFFIX = ".error"
RECENT_DAYS = 40
ONE_MONTH = relativedelta(months=1)
NB_WORKERS=10

class PathInfo :
    def __init__(self):
        self.path = None
        self.recent = False

def date_placeholders(date) :
    """ Generate a dict of placeholder for start / end dates : YYYY MM DD / YYYYe MMe DDe """

    def date_dict(date, suffix = "") :
        return {
            "YYYY" + suffix : date.strftime("%Y"),
            "MM" + suffix : date.strftime("%m"),
            "DD" + suffix : date.strftime("%d")}

    return {
        **date_dict(date),
        **date_dict(date+ONE_MONTH, "e")}

def list_downloads(properties, url_pattern, path_pattern, start_date=None, end_date=None) :

    res = defaultdict(lambda : PathInfo())

    properties = dict((key, parse_value(val)) for key, val in properties.items())

    # No end date ? => until now
    if not end_date :
        end_date_str = properties.get("EndDate", None)
        end_date = datetime.now() if end_date_str is None else datetime.strptime(end_date_str, DATE_FORMAT)


    # Loop on months
    if not start_date :
        start_date =  datetime.strptime(properties["StartDate"], DATE_FORMAT)

    # Start first of the month
    date =  start_date.replace(day=1)

    while date <= end_date:

        date_dict = date_placeholders(date)

        url = url_pattern.format(**properties, **date_dict)
        path = path_pattern.format(**properties, **date_dict)

        res[url].path = path

        # "Recent" chunk ?
        if datetime.now() - date < timedelta(days=RECENT_DAYS) :
            res[url].recent = True

        date += ONE_MONTH

    return res

def zip_file(inf, outf) :
    with open(inf, 'rb') as f_in:
        with gzip.open(outf, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)



def do_download(url_paths, out, dry_run=False, compress=False) :

    def process_one(args):
        url, pathInfo = args

        out_path = os.path.join(out, pathInfo.path)

        with LogContext(file=out_path), IgnoreAndLogExceptions():

            # Skip file if already present, unless it is "recent"
            if os.path.exists(out_path) or os.path.exists(out_path + ERROR_SUFFIX):

                if pathInfo.recent:
                    info("File %s is already present but recent. Check if newer version exists", out_path)
                else:
                    info("File %s is already present. Skipping", out_path)
                    return

            folder = os.path.dirname(out_path)
            if not os.path.exists(folder):
                os.makedirs(folder)

            with NamedTemporaryFile() as tmpFile:

                if dry_run:
                    info("Would have downloaded %s -> %s " + ("[compressed]" if compress else ""), url, out_path)
                else:
                    info("Downloading %s -> %s", url, out_path)
                    urlretrieve(url, tmpFile.name)

                    if compress:
                        zip_file(tmpFile.name, out_path)
                    else:
                        # Files already exists ? Only update if size are different
                        if os.path.exists(out_path) and os.path.getsize(out_path) == os.path.getsize(tmpFile.name):
                            info("File {} was already present with same size => skipping")
                        else:
                            shutil.copy(tmpFile.name, out_path)

    # Parallel execution : wait for all executions to finish
    with ThreadPoolExecutor(max_workers=NB_WORKERS) as executor:
        executor.map(process_one, url_paths.items())



def parse_date(s):
    return datetime.strptime(s, '%Y-%m-%d')

def main() :

    networks_info = getNetworksInfo()

    parser = argparse.ArgumentParser(description='Get raw data files from HTTP APIs')
    parser.add_argument('network', metavar='<network>', choices=list(networks_info.keys()), help='Network')
    parser.add_argument('out_folder', metavar='<dir>', type=str, help='Output folder')
    parser.add_argument('--ids', metavar='station_id1,station_id2', type=str, help='Optional IDs', default=None)
    parser.add_argument('--start-date', metavar='yyyy-mm-dd', type=parse_date, help='Start date, optional (start of station by default)', default=None)
    parser.add_argument('--end-date', metavar='yyyy-mm-dd', type=parse_date, help='End date, optional (end of station by default)', default=None)
    parser.add_argument('--dry-run', '-n', action='store_true', help='Do not download anything. Only print what would be downloaded')
    args = parser.parse_args()

    stations = getStationsInfo(args.network)

    network_info = networks_info[args.network]
    url_pattern = network_info[SOURCE_URL_ATTR]
    path_pattern = network_info[RAW_PATH_ATTR]
    compress = parse_bool(network_info[COMPRESS_ATTR])

    if not url_pattern :
        raise Exception("'SourceURL' not defined for network %s" % args.network)

    station_ids = None if args.ids is None else args.ids.split(",")

    for id, properties in stations.items():

        if station_ids and not id in station_ids :
            continue

        with LogContext(network=args.network, station_id=id) :

            url_paths = list_downloads(properties, url_pattern, path_pattern, args.start_date, args.end_date)
            do_download(url_paths, args.out_folder, args.dry_run, compress)

if __name__ == '__main__':
    main()






import re
from copy import deepcopy
from typing import Dict
from libinsitu import STATION_NAME_VAR
from libinsitu.common import parse_value, DATA_VARS, read_res, CDL_PATH, LONGITUDE_VAR, LATITUDE_VAR, ELEVATION_VAR, \
    fill_str, TIME_VAR
from libinsitu.log import info, warning


SYSTEM_ATTRIBUTES = ["_FillValue"]

class Variable :
    def __init__(self, name, type, dimensions):
        self.type = type
        self.dimensions = dimensions
        self.name = name
        self.attributes = {}

class CDL :
    def __init__(self):
        self.dimensions : Dict[str, int] = {}
        self.variables : Dict[str, Variable] = {}
        self.global_attributes = {}

def replace_placeholders(strval, attributes) :

    def repl(m) :
        key = m.group().strip("{").strip("}")

        if not key in attributes :
            warning("Key : '%s' not found in attributes, using empty string instead" % key)
            return ""

        res = attributes[key]
        return "" if res is None else str(res)

    return re.sub(r'{\w+}', repl, strval)

def parse_cdl(lines, attributes=dict()) :
    """ Parse CDL file """

    res = CDL()

    section = None
    curr_var = None

    for line in lines :
        line = line.strip()

        # Skip comments
        if line.startswith("#") or len(line) == 0:
            continue

        if "#" in line :
            line = line.split("#")[0]
            line = line.strip()

        # key = value
        if "=" in line :
            line = line.strip(";")
            key, val = line.split("=", 1)
            key = key.strip()
            val = parse_value(val.strip(), split=True)

            if isinstance(val, str):
                val = replace_placeholders(val.strip(), attributes)

            # Defining dimension
            if section == "dimensions" :

                dim = 0 if val == "UNLIMITED" else int(val)
                res.dimensions[key] = dim

            elif section == "variables" :
                varname, attrname = key.split(":")

                if varname == "*" :
                    varname = curr_var

                if varname == "" :
                    res.global_attributes[attrname] = val
                else :
                    res.variables[varname].attributes[attrname] = val
            else :
                raise Exception("Assignement outside any section : %s" % line)


        # Skip start or end
        elif "{" in line or "}" in line :
            continue

        # Change section
        elif ":" in line :
            section = line.strip(":").strip()
            continue

        elif ";" in line :
            # New var
            line = line.strip(";").strip()
            type, var = line.split()

            if type == "char" :
                type ="c"
            elif type == "string" :
                type = str
            elif type == "float" :
                type="f4"
            elif type == "int" :
                type="i4"
            elif type == "short":
                type = "i2"
            elif type == "uint":
                type = "u4"

            dims=[]
            if "(" in var :
                var, dims = var.split("(")
                dims = dims.strip(")").strip().split(",")
            res.variables[var] = Variable(var, type, dims)
            curr_var = var
        else :
            raise Exception("Bad line : %s" %line)

    return res

def update_attributes(dest, src, dry_run=False, delete=False) :

    dry_prefix = "would " if dry_run else ""

    existing_attrs = set(dest.ncattrs())
    new_attrs = set(src.keys())

    extra_attrs = existing_attrs - new_attrs

    if delete :
        for attrname in extra_attrs :
            info(dry_prefix + "delete attribute %s#%s" % (dest.name, attrname))
            if not dry_run :
                dest.delncattr(attrname)

    for key, val in src.items() :
        oldval = None if not key in existing_attrs else dest.getncattr(key)

        if key in SYSTEM_ATTRIBUTES :
            # Do not update system attributes
            continue

        if oldval != val :

            if (val is None or val == "") and not delete :
                continue

            info(dry_prefix + "update attribute %s#%s %s -> %s" % (dest.name, key, oldval, val))

            if not dry_run:
                dest.setncattr(key, val)


def cmp_var(var,  vardef:Variable) :
    return var.dtype == vardef.type and var.dimensions == tuple(vardef.dimensions)

def create_or_replace_var(ncfile, vardef:Variable, dry_run=False) :

    if vardef.name in ncfile.variables:

        var = ncfile.variables[vardef.name]

        if cmp_var(var, vardef):
            # Same var -> skipping
            return

        # Different vars
        if dry_run:
            info("Would replace var : %s" % vardef.name)
        else:
            info("Replacing var : %s" % vardef.name)
            del ncfile.variables[vardef.name]

    if dry_run :
        info("Would add variable : %s" % vardef.name)
        return

    least_significant_digit = vardef.attributes.get("least_significant_digit", None)
    fill_value = vardef.attributes.get("_FillValue", None)

    info("Adding variable '%s'. Precision:%s" %  (vardef.name, least_significant_digit))

    ncfile.createVariable(
        vardef.name, vardef.type, vardef.dimensions,
        zlib=vardef.type is not str,
        complevel=9,
        least_significant_digit=least_significant_digit,
        fill_value=fill_value)

def initVar(ncfile, vardef:Variable, dry_run=False, delete_attrs=False) :

    create_or_replace_var(ncfile, vardef, dry_run)

    var = ncfile.variables[vardef.name]

    # Update attributes
    update_attributes(var, vardef.attributes, dry_run, delete_attrs)

def cdl2netcdf(ncfile, cdl: CDL, dry_run=False, delete_attrs=False) :
    """Init NetCDF file from a CDL"""

    for dimname, dim in cdl.dimensions.items() :

        # Already there, skipping
        if not dimname in ncfile.dimensions and not dry_run :
            info("Adding dimension '%s'", dimname)
            ncfile.createDimension(dimname, dim)

    for varname, vardef in cdl.variables.items() :
        initVar(ncfile, vardef, dry_run, delete_attrs)

    # Update global attributes
    update_attributes(ncfile, cdl.global_attributes, dry_run, delete_attrs)


def init_nc(netcdf, properties, data_vars=DATA_VARS, dry_run=False, delete_attrs=False) :

    try:
        # Try to load custom CDL first
        custom_name = properties["Network_ID"] + ".cdl"
        cdl = parse_cdl(read_res(custom_name), properties)
        info("Used custom CDL : %s" % custom_name)

    except FileNotFoundError:
        cdl = parse_cdl(read_res(CDL_PATH), properties)

    # Ensures all requested data vars are defined
    missing_vars = set(data_var for data_var in data_vars if data_var not in cdl.variables)
    if len(missing_vars) > 0 :
        raise Exception("Unknown data vars : %s" % missing_vars)

    filtered_cdl = deepcopy(cdl)

    # Filter data vars (variables with "time" dimension)
    # Also adds the "Time" variable
    filtered_cdl.variables = dict()
    for key, var in cdl.variables.items() :
        if "time" in var.dimensions and not key in data_vars + [TIME_VAR] :
            continue
        filtered_cdl.variables[key] = var

    cdl2netcdf(netcdf, filtered_cdl, dry_run, delete_attrs)

    if not dry_run :

        # Init scalar vars
        netcdf.variables[LONGITUDE_VAR][0] = properties["Station_Longitude"]
        netcdf.variables[LATITUDE_VAR][0] = properties["Station_Latitude"]
        netcdf.variables[ELEVATION_VAR][0] = properties["Station_Elevation"]

        fill_str(netcdf, STATION_NAME_VAR, properties["Station_ID"])

    return cdl







#!/usr/bin/env python
import csv
import json
import os, sys

import argparse
from rich.console import Console
from rich.table import Table

from libinsitu.catalog import *
from libinsitu.log import console


def printCatalog(catalog, rec=False) :

    console = Console()

    if len(catalog.catalogs) > 0 :

        print("Sub catalogs : ")

        table = Table(show_header=True)
        table.add_column("Name", style="dim", width=12)
        table.add_column("url")

        for key, cat in catalog.catalogs.items() :
            table.add_row(
                cat.name,
                cat.url)

        console.print(table)

    if len(catalog.datasets) > 0 :

        print("Datasets : ")

        table = Table(show_header=True, show_lines=True)
        table.add_column("Name", style="dim", width=12)
        table.add_column("Protocol", style="dim", width=12)
        table.add_column("url")

        for key, dataset in catalog.datasets.items():
            table.add_row(
                dataset.name,
                "\n".join(dataset.services.keys()),
                "\n".join(dataset.services.values()))

        console.print(table)

    if rec :
        for sub_cat in catalog.catalogs.values() :
            printCatalog(sub_cat)

def filter_services(catalog, services) :
    for ds in catalog.datasets.values() :
        ds.services = {serv : url for serv, url in ds.services.items() if serv in services}
    for cat in catalog.catalogs.values() :
        filter_services(cat, services)

def flatten_ds(cat) :

    def list_ds(cat) :
        res = {}
        res.update(cat.datasets)
        for subcat in cat.catalogs.values() :
            res.update(list_ds(subcat))
        return res

    cat.datasets = list_ds(cat)
    cat.catalogs = {}

def output_csv(catalog) :
    writer = csv.DictWriter(sys.stdout, ["id", "name", "protocol", "url"])
    writer.writeheader()
    def dump_services(catalog) :
        for ds in catalog.datasets.values() :
            for protocol, url in ds.services.items() :
                writer.writerow(dict(
                    id = ds.id,
                    name = ds.name,
                    protocol = protocol,
                    url = url
                ))
        for cat in catalog.catalogs.values() :
            dump_services(cat)

    dump_services(catalog)


def main():

    # Put logs to stderr
    console.file = sys.stderr

    parser = argparse.ArgumentParser(description='Browse a TDS (THREDDS) catalog')
    parser.add_argument('url', metavar='<http://host/catalog.xml>', type=str, help='Start URL')
    parser.add_argument('--user', '-u', help='User login (or TDS_USER env var)',
                        default=os.environ.get("TDS_USER", None))
    parser.add_argument('--password', '-p', help='User password (or TDS_PASS env var)',
                        default=os.environ.get("TDS_PASS", None))
    parser.add_argument('--recursive', '-r', action="store_true", help="Fetch recursively")
    parser.add_argument('--output-format', '-of', choices=["display", "json", "csv"], help="Output format", default="display")
    parser.add_argument('--flatten', '-fl', action="store_true", help="Flatten datasets")
    parser.add_argument('--services', '-s', metavar="serv1,serv2, ...", help="Filter on list of services.")

    args = parser.parse_args()

    session = Session()
    if args.user:
        session.auth = (args.user, args.password)

    catalog = fetch_catalog(args.url, session, recursive=args.recursive)

    if args.services :
        # Filter only some services
        filter_services(catalog, args.services.split(","))

    if args.flatten :
        flatten_ds(catalog)

    if args.output_format == "json" :
        json.dump(
            catalog, sys.stdout,
            default=lambda d: {k : v for k, v in d.__dict__.items()},
            indent=2)
    elif args.output_format == "display" :
        printCatalog(catalog, rec=args.recursive)
    elif args.output_format == "csv" :
        output_csv(catalog)
    else:
        raise Exception("Format not supported : %s" % args.output_format)


if __name__ == '__main__':
    main()





from libinsitu.common import GLOBAL_VAR, DIRECT_VAR, DIFFUSE_VAR
from libinsitu.handlers.base_handler import InSituHandler, map_cols
import pandas as pd

# The time base for all readings is South African Standard Time (SAST" \
# See : http://www.scielo.org.za/scielo.php?script=sci_arttext&pid=S1021-447X2015000100001
TIMEZONE=2

class SAURANHandler(InSituHandler) :

    def _read_chunk(self, stream) :

        GHI_Col = self.properties["Station_GHI_Col"]
        DHI_Col = self.properties["Station_DHI_Col"]
        DNI_Col = self.properties["Station_DNI_Col"]

        mapping = {
           DHI_Col: DIFFUSE_VAR,
           DNI_Col: DIRECT_VAR,
           GHI_Col: GLOBAL_VAR}

        data = pd.read_csv(
            stream,
            skiprows=[0, 2, 3],
            parse_dates=["TmStamp"], index_col="TmStamp", dayfirst=True,
            usecols=["TmStamp"] + [GHI_Col, DHI_Col, DNI_Col])


        data = map_cols(data, mapping)

        # Shift Timezone
        data.index -= pd.to_timedelta(TIMEZONE, "H")

        return data


    def data_vars(self):
        """ @override """
        return [GLOBAL_VAR, DIFFUSE_VAR, DIRECT_VAR]





from pvlib.iotools import parse_bsrn
from libinsitu.common import GLOBAL_VAR, DIRECT_VAR, DIFFUSE_VAR, TEMP_VAR, HUMIDITY_VAR, PRESSURE_VAR, DATA_VARS
from libinsitu.handlers.base_handler import InSituHandler, map_cols, ZERO_DEG_K
from libinsitu.log import error
import pandas as pd

MAPPING = dict(
    ghi=GLOBAL_VAR,
    dni=DIRECT_VAR,
    dhi=DIFFUSE_VAR,
    temp_air=TEMP_VAR,
    relative_humidity=HUMIDITY_VAR,
    pressure=PRESSURE_VAR)

class BSRNHandler(InSituHandler) :

    def _read_chunk(self, stream) :

        data, metadata = parse_bsrn(stream)

        data = map_cols(data, MAPPING)

        # Check type of column
        for col in self.data_vars() :
            if data[col].dtype == object :
                # String ? A couple of values might be incorrent.
                # Try to convert to float, ignoring errors
                error("Column %s parsed as String : converting to float. errors will be NaN", col)
                data[col] = pd.to_numeric(data[col], errors="coerce")

        # Convertions
        data[TEMP_VAR] = data[TEMP_VAR] + ZERO_DEG_K # T2: °C -> K
        data[HUMIDITY_VAR] = data[HUMIDITY_VAR] / 100  # percent -> 1
        data[PRESSURE_VAR] = data[PRESSURE_VAR] * 100 # Pressure hPa->Pa

        return data

    def data_vars(self):
        """ @override """
        return [GLOBAL_VAR, DIFFUSE_VAR, DIRECT_VAR, TEMP_VAR, HUMIDITY_VAR, PRESSURE_VAR]


import os.path
import re
from abc import abstractmethod
from datetime import datetime
from glob import glob
from gzip import GzipFile
from io import TextIOWrapper
from pathlib import PurePath
from typing import Dict
from zipfile import ZipFile

from pandas import DataFrame

from libinsitu import match_pattern
from libinsitu.log import warning, debug

ZERO_DEG_K = 273.15

def map_cols(data, mapping) :
    """Filter and rename columns """
    data = data[list(mapping.keys())]
    return data.rename(columns=mapping)

class InSituHandler :
    """ Virtual class to be implemented for each new network """
    
    def __init__(self, properties):
        self.properties = properties.copy()

        # Also adds lower case version of properties
        for key, val in properties.items() :
            if isinstance(val, str) :
                self.properties[key.lower()] = val.lower()


    # final
    def read_chunk(self, filename:str, encoding='latin1'):
        """ Handle opening of gz / zip files. Filename can contain a zip entry name after '!' """

        zip_entry= None
        if '!' in filename :
            filename, zip_entry = filename.split('!')

        if filename.endswith(".gz") :
            with open(filename, "rb") as f:
                stream =  TextIOWrapper(GzipFile(fileobj=f), encoding=encoding)
                return self._read_chunk(stream)

        elif filename.endswith('.zip'):  # check if file is a zipped (.zip) file

            with ZipFile(filename) as thezip :

                names = thezip.namelist()

                if len(names) == 1 :
                    entry = names[0]
                elif zip_entry is not None: # Explicit zip entry request after !
                    if not zip_entry in names :
                        raise Exception("Missing zip entry '%s' in '%s'" % (zip_entry, filename))
                    entry = zip_entry
                else:
                    # TODO : This was a default harcoded : should be better to set it in each subclass
                    entry = [tmp for tmp in thezip.namelist() if '.txt' in tmp][0]

                stream = thezip.open(entry, mode="r")
                return self._read_chunk(stream)

        else :
            with open(filename, "rt", encoding=encoding) as f :
                return self._read_chunk(f)

    @abstractmethod
    def pattern(self):
        """Should return a file pattern for input files.

        The following placeholders are supported :
        *: any string
        ?: any single caracter
        {MM} : Month of chunk file
        {YYYY} / {YY} : Year of chunk file
        {DDD} : Day of year (1-365)
        {Property_Name} : Any property defined in station-info csv file
        {property_name} : Same property, in lower case

        The pattern is used to sort file by year and month.
        If not provided, the year and month of the modification of the file are used.

        Example patterns :
        - "{Station_ID}-{YY}-{MM}*.zip"
        - "???{ID}*.txt"
        """

        # Take this from the RawDataPath property of the network
        return self.properties["Network_RawDataPath"]

    def glob_pattern(self) :
        """Transforms the pattern to a glob pattern"""
        def subf(match) :
            key = match.group(0).replace("{", "").replace("}", "")
            if key in self.properties:
                return str(self.properties[key])
            elif key in ["YY", "M", "MM", "YYYY", "DDD"] :
                return "?" * len(key)
            else :
                raise Exception("Unsupported pattern : %s" % key)

        return re.sub(r'\{\w+\}', subf, self.pattern())



    def list_files(self, folder):
        """List files from folder matching the pattern """

        # First go a glob
        pattern = folder + "/" + self.glob_pattern()

        debug("Pattern :", pattern)

        filenames = list(_zip_glob(pattern))

        debug("Candidate filenames :", filenames)

        # Finer filter on each name
        def filter_f(filename) :
            basename = os.path.basename(filename)
            return self.match_pattern(basename) is not False

        res =  list(filename for filename in filenames if filter_f(filename))

        debug("Filtered filenames :", res)

        return res

    def match_pattern(self, value):
        pattern = os.path.basename(self.pattern())
        return match_pattern(pattern, value, self.properties)

    def sort_files(self, filenames):

        def sort_key(filename) :
            basename = os.path.basename(filename)

            match_groups = self.match_pattern(basename)
            if not match_groups :
                warning("File %s does not match pattern %s. It may not not be included in correct order" % (basename, self.pattern()))
                return basename

            # By default, use year and month of modification time
            realfilename = filename
            if '!' in filename :
                realfilename, zip_entry = filename.split("!")
            mtime = datetime.fromtimestamp(os.path.getmtime(realfilename))
            year = mtime.year
            month_or_days = mtime.month


            if "M" in match_groups :
                month_or_days = int(match_groups["M"])
            if "MM" in match_groups:
                month_or_days = int(match_groups["MM"])
            if "DDD" in match_groups:
                month_or_days = int(match_groups["DDD"])
            if "YYYY" in match_groups :
                year = int(match_groups["YYYY"])
            if "YY" in match_groups :
                year = int(match_groups["YY"])
                year = year + (2000 if year < 70 else 1900)

            return (year, month_or_days, basename)

        filenames_keys = { filename: sort_key(filename) for filename in filenames}

        return sorted(list(filenames_keys.keys()), key=lambda filename : filenames_keys[filename])


    def data_vars(self):
        # By default, return the list of available data from the CSV files
        return self.properties["Network_AvailableData"].split(",")

    @abstractmethod
    def _read_chunk(self, stream) -> DataFrame:
        """
        Should return a panda DataFrame, with a datetime index and columns correponding to #DATA_VARIABLES, as defined in common.py.
        Missing values should be np.nan
        """
        pass

def _zip_entries(zip_file) :
    with ZipFile(zip_file) as thezip:
        return thezip.namelist()

def _zip_glob(pattern) :
    """Extension of 'glob' that supports looking into ZIP file entries (after '!')"""
    entries_pattern = None
    if '!' in pattern :
        pattern, entries_pattern = pattern.split("!")

    files = glob(pattern)
    if entries_pattern is None :
        return files

    res = []
    for zipfile in files :
        for entry in  _zip_entries(zipfile) :
            if PurePath(entry).match(entries_pattern) :
                res.append(zipfile + '!' + entry)
    return res






from libinsitu.common import GLOBAL_VAR, DIFFUSE_VAR, DIRECT_VAR, TEMP_VAR, WIND_SPEED_VAR, WIND_DIRECTION_VAR
from libinsitu.handlers.base_handler import map_cols, InSituHandler, ZERO_DEG_K

from pvlib.iotools import surfrad


class RADHandler(InSituHandler) :

    def read_chunk(self, filename:str, encoding='latin1'):
        """ @override """
        data, metadata = surfrad.read_surfrad(filename, map_variables=False)
        mapping = {
            "dw_solar" :  GLOBAL_VAR,
            "direct_n" : DIRECT_VAR,
            "diffuse" : DIFFUSE_VAR,
            "temp" : TEMP_VAR,
            "windspd" :  WIND_SPEED_VAR,
            "winddir" : WIND_DIRECTION_VAR}

        data = map_cols(data, mapping)

        # Conversions
        data[TEMP_VAR] = data[TEMP_VAR] + ZERO_DEG_K

        return data

    def data_vars(self):
        """ @override """
        return [GLOBAL_VAR, DIFFUSE_VAR, DIRECT_VAR, TEMP_VAR, WIND_SPEED_VAR, WIND_DIRECTION_VAR]



from jupyter_dash import JupyterDash
from dash import Dash, html, dcc
import pandas as pd

from libinsitu import visual_qc


def get_data():

    # For more details on this function, check : https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
    df = pd.read_csv(
        "data/sample-data.csv", # Input filename
        usecols=["date", "time", "global", "diffuse", "direct"], # Only get the columns we need (drop temp)
        parse_dates = [["date", "time"]], # We can combine several columns to build timestamp
        index_col = "date_time") # Use the parsed date time as index

    # Rename columns
    df.rename(inplace=True, columns={"global":"GHI", "diffuse":"DHI", "direct":"BNI"})

    # Metadata of the station
    df.attrs["latitude"] = 39.742
    df.attrs["longitude"] = -105.18
    df.attrs["elevation"] = 1828.8
    df.attrs["station_name"] = "BMS"

    return df

if __name__ == '__main__':
    app = JupyterDash(__name__)

    df = get_data()

    fig = visual_qc(df, engine="plotly")

    app.layout = dcc.Graph("timeseries", fig)

    app.run_server(mode='inline', port=8050)


import argparse
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from multiprocessing import Lock

import sg2
from netCDF4 import Dataset

from libinsitu import read_res, info, netcdf_to_dataframe, LATITUDE_VAR, LONGITUDE_VAR, ELEVATION_VAR, STATION_NAME_VAR, \
    datetime64_to_sec, sec_to_datetime64, getTimeVar, TIME_DIM, QC_FLAGS_VAR, qc_masks
from libinsitu.cdl import cdl2netcdf, parse_cdl
import pandas as pd
import numpy as np

from libinsitu.log import LogContext

INDEX_CDL = "index.cdl"

WRITE_LOCK = Lock()

EXPECTED_COUNT_VAR = "expected_daylight_count"
VALID_COUNT_SUFFIX = "_valid_daylight_count"
QC_COUNT_PATTTERN = QC_FLAGS_VAR + "_%s_daylight_count"

def parser() :

    parser = argparse.ArgumentParser(description='Produces daily index NetCDF files from other files')
    parser.add_argument('output', metavar='<out.nc>', help="Output NetCDF file")
    parser.add_argument('inputs', metavar='<file.nc>', help="Input NetCDF files", nargs="+")
    parser.add_argument('--max-threads', metavar='<nb_threads>', type=int, default=None)

    return parser


def main() :

    args = parser().parse_args()

    out_nc = Dataset(args.output, "w")

    properties = dict()

    # Init output NetCDF
    properties["NbStations"] = len(args.inputs)
    cdl = parse_cdl(read_res(INDEX_CDL), properties)
    cdl2netcdf(out_nc, cdl)

    # Open all input NetCDF (not expensive, since no data is read at this point)
    input_ncs = list(Dataset(input, "r") for input in args.inputs)

    # Gather all data vars
    data_vars = set()
    for input_nc in input_ncs :
        data_vars = data_vars.union(list_data_vars(input_nc))

    # Create all vars upfront
    create_count_var(out_nc, EXPECTED_COUNT_VAR)
    for varname in data_vars:
        create_count_var(out_nc, varname + VALID_COUNT_SUFFIX)

    # Init QC flags if any
    for input_nc in input_ncs :
        if QC_FLAGS_VAR in input_nc.variables:
            qc_var = input_nc.variables[QC_FLAGS_VAR]
            for flag in qc_var.flag_meanings.split() :
                create_count_var(out_nc, QC_COUNT_PATTTERN % flag)
            break

    start_time = min_time(input_ncs)
    start_day = datetime64_to_sec(out_nc, start_time, 'D')

    # Close output to allow multi process
    out_nc.close()

    # Parallel execution
    executor = ProcessPoolExecutor(max_workers=args.max_threads)
    res = executor.map(
        partial(process_station, start_day, args.output),
        range(len(args.inputs)), # istation
        args.inputs) # infile
    list(res)


def min_time(input_ncs) :
    res= None
    for input in input_ncs :
        start_date = sec_to_datetime64(input, getTimeVar(input)[0])
        if res is None or res > start_date :
            res = start_date
    return res

def process_station(start_day, outfile, istation, infile) :

    with LogContext(file=infile) :

        in_dfs = netcdf_to_dataframe(infile, chunked=True, chunk_size=1000000)

        # Use chunked processing to reduce memory usage
        for ichunk, in_df in enumerate(in_dfs):

            chunk_id = "#%d %s -> %s" % (ichunk, in_df.index.min(), in_df.index.max())

            info("Processing %s. Chunk %s" % (infile, chunk_id))

            # First compute data and then write it all at once
            data_dic = dict()

            lat = in_df.attrs[LATITUDE_VAR]
            lon = in_df.attrs[LONGITUDE_VAR]
            alt = in_df.attrs[ELEVATION_VAR]
            name = in_df.attrs[STATION_NAME_VAR]

            # Compute a boolean index of daylight records
            toa = compute_toa(lat, lon, alt, in_df.index)
            is_daylight = toa.TOA > 0

            # Daily expected number of records
            expected_daylight_count = is_daylight.resample('D').sum()

            data_dic[EXPECTED_COUNT_VAR] = expected_daylight_count

            for col in in_df.columns :

                if col == QC_FLAGS_VAR :
                    continue

                #info("Processing for variable %s %s" % (col, "(with QC)" if QC_FLAGS_VAR in in_df else ""))

                not_na = ~in_df[col].isna()

                valid_daylight = not_na & is_daylight

                if QC_FLAGS_VAR in in_df :
                    valid_daylight = valid_daylight & (in_df.QC == 0)

                valid_daily = valid_daylight.resample('D').sum()

                #write_series(out_nc, istation, col + "_valid_daylight_count", valid_daily, start_day)
                data_dic[col + VALID_COUNT_SUFFIX] = valid_daily

            # QC
            if QC_FLAGS_VAR in in_df :
                qc_col = in_df[QC_FLAGS_VAR]
                for flag, mask in qc_masks(in_df).items() :
                    data_dic[QC_COUNT_PATTTERN % flag] = is_daylight & ((qc_col & mask) != 0)

            with WRITE_LOCK :

                info("Writing output for chunk %s" % chunk_id)

                out_nc = Dataset(outfile, mode="a")
                try:

                    # Write meta data
                    out_nc.variables[LATITUDE_VAR][istation] = lat
                    out_nc.variables[LONGITUDE_VAR][istation] = lon
                    out_nc.variables[ELEVATION_VAR][istation] = alt
                    out_nc.variables[STATION_NAME_VAR][istation] = name

                    # write data
                    for key, data in data_dic.items() :
                        write_series(out_nc, istation, key, data, start_day)
                finally:
                    out_nc.close()




def list_data_vars(ncfile) :
    res = set()
    timeVar = getTimeVar(ncfile)
    for varname, var in ncfile.variables.items() :
        if TIME_DIM in var.dimensions and var != timeVar and varname != QC_FLAGS_VAR :
            res.add(varname)
    return res

def create_count_var(ncfile, name):
        info("Adding var : %s" % name)
        ncfile.createVariable(
            name, int, ["station", "time"],
            zlib=True,
            complevel=9,
            fill_value=-1)

def write_series(out_nc, istation, var_name, series, ref_day) :

    time_var = getTimeVar(out_nc)

    # Fill holes / make time series regular
    series = series.asfreq("1D")

    nb_times = len(time_var)

    start_day = datetime64_to_sec(out_nc, series.index.min(), 'D')
    end_day = datetime64_to_sec(out_nc, series.index.max(), 'D')
    end_idx = end_day - ref_day

    if end_idx >= nb_times :
        time_var[0:end_idx+1] = np.arange(ref_day, end_day+1)



    out_var = out_nc.variables[var_name]

    start_idx = start_day - ref_day

    out_var[istation, start_idx:] = series.values


def compute_toa(lat, lon, alt, times) :

    sun_pos = sg2.sun_position(
        [[lon, lat, alt]],
        times,
        ["topoc.toa_hi"])

    df = pd.DataFrame(dict(
        TOA=np.squeeze(sun_pos.topoc.toa_hi)),
        index=times)

    return df
